//////////////////////////////////////////////////////////////////////////////// 
// 
// history:
//
// 2020-05-01  uj  0.1.0  first setup
// 2021-06-04  uj  1.1.0  handle empty cells when reading XLS
// 2022-01-04  uj  2.0.0  major extensions for SECTOR
// 2022-07-14  uj  2.1.0  handle negative modified values in the FOI data
// 2023-07-04  uj  2.1.1  handle empty UID string in XLS EEDB
//
////////////////////////////////////////////////////////////////////////////////

/**
 * (c) Janicke Consulting, Germany 2023
 * 
 * EEDB: Engine Emission Data Base management
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

package de.janicke.airport.eedb;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.util.HashMap;
import java.util.Locale;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.TimeZone;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellType;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.xssf.usermodel.XSSFCell;
import org.apache.poi.xssf.usermodel.XSSFRow;
import org.apache.poi.xssf.usermodel.XSSFSheet;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

public class EEDB {
  
  public static final String VERSION = "2.1.1";
  public static boolean DEBUG = false;
  public boolean CHECK = true;
  
  public static enum ACGT { Large, Medium, Small, Regional, Business, Turboprop, Piston, HeliLarge, HeliSmall }; 
  public static enum SOURCE { ICAO, FOCA, FOI, OTHER, ACG, UNDEF };
  public static enum DOMAIN { INPUT, EMIS };
  public static enum GROUP { AC, APU };
  public static enum APUM { SS, HL, NR };
  public static enum APUT { A997, A996, A995, A994, A993, A992, A991 };
  
  public static final double FSCDEFAULT = 0.00068;
  public static final double S4TOS6DEFAULT = 0.024;
  public static final double H2ODEFAULT = 0.14;

  public static final double EI_SOX = 0.8;
  public static final double EI_FB = 1000.0;
  public static final double EI_CO2 = 3155.0;
  public static final double BNZ_IN_HC = 0.02;         // factor BNZ/HC in g/g
  public static final double ODOR_IN_HC = 31000.;      // factor ODOR/HC in OU/g
  public static final double PM25_IN_PM10 = 1.;        // factor PM25/PM10 in g/g

  public static final String MEASURED = "MEASURED";
  public static final String UNDEF = "UNDEF";
  public static final String UNIT = "UNIT";
  public static final String TOTAL = "TOTAL";
  public static final String UNKNOWN = "Unknown";
  public static final String APU = "APU";
  public static final String AC = "AC";
  public static final String TIM = "TIM";
  public static final String FUELFLOW = "FF";
  public static final String FFWEIGHT = "FFW";
  public static final String FUELBURN = "FB";
  public static final String SMOKENUMBER = "SN";
  public static final String HC = "HC";
  public static final String CO = "CO";
  public static final String NOX = "NOX";
  public static final String NO2 = "NO2";
  public static final String NVPN = "NVPN";
  public static final String NVPM = "NVPM";
  public static final String X_UNCORRECTED = "_UC";
  public static final String NVPN_UC = NVPN+X_UNCORRECTED;
  public static final String NVPM_UC = NVPM+X_UNCORRECTED;
  public static final String PM10 = "PM10";
  public static final String PM25 = "PM25";
  public static final String PM10_NV = "PM10_NV";
  public static final String PM10_VS = "PM10_VS";    // deprecated
  public static final String PM10_SO4 = "PM10_SO4";
  public static final String PM10_VH = "PM10_VH";   // deprecated
  public static final String PM10_VO = "PM10_VO";
  public static final String PM10_POM = "PM10_POM";
  public static final String VPN = "VPN";
  public static final String VPM = "VPM";
  public static final String TPN = "PN";
  public static final String TPM = "PM";
  public static final String PMR = "PMR";
  public static final String LTO = "LTO";
  public static final String SNMAX = "SNMAX";
  public static final String BNZ = "BNZ";
  public static final String ODOR = "ODOR";
  public static final String CO2 = "CO2";
  public static final String SOX = "SOX";
  public static final String GMD = "GMD";
  public static final String CEE = "CEE"; 
  
  public static final String[] EEDBTRCORG = { "HC",  "CO", "NOX",        "SN",   "Fuel", "EImass ", "EInum ", "EImass_SL", "EInum_SL" };
  public static final String[] EEDBTRCLSP = {   HC,    CO,   NOX, SMOKENUMBER, FUELFLOW,   NVPM_UC,  NVPN_UC,        NVPM,      NVPN  };
  public static final double[] EEDB2LSP =   {   1.,    1.,    1.,          1.,       1.,     1.e-3,       1.,       1.e-3,        1.  };
  
  public static final String[] FOCATRCORG = { "VOC",  "CO", "NOX",     "FF",  "PMF", "nvPMF",  "nvPNF" };
  public static final String[] FOCATRCLSP = {    HC,    CO,   NOX, FUELFLOW,   PM10,    NVPM,     NVPN };
  public static final double[] FOCA2LSP =   {    1.,    1.,    1.,       1.,     1.,      1.,       1. };
  
  public static final String[] FOITRCORG = {       "Fuel", "HC",  "CO",  "NOX"  };
  public static final String[] FOITRCLSP = {     FUELFLOW,   HC,    CO,    NOX  };
  public static final double[] FOI2LSP =   {           1.,   1.,    1.,      1. };
  
  public static final String[] CSVTRC = {   FUELFLOW,   HC,    CO,     NOX, SMOKENUMBER,   PM10,  NVPM,    NVPN  };
  
  public static final char SEPARATOR = ';';
  public static final char SEPREPLACE = ',';
  public static final String OUTEXCLUDE = "[^a-zA-Z0-9()_/\\s\\-.,\\[\\]]";
  
  private LinkedHashMap<String,ENG> hm;
  private LinkedHashMap<String,ACG> hg;
  private LinkedHashMap<String,Double> hapu;
  private LinkedHashMap<String,PRM> hp;

  /**
   * Class to handle parameters
   */
  private class PRM extends Object {
    String name;
    Class type;
    Object value;
    
    public PRM(String name, Class cl, Object v) {
      this.name = name;
      this.type = cl;
      this.value = v;
    }
  }
  
  /**
   * Class to handle table columns
   */
  private class ColumnIndex {
    //
    // EEDB column headers
    private String[] EEDBMOD = { "t/o", "c/o", "app", "idle", " lto " };
    public final String sUID = "uid no";
    public final String sENG = "engine identification";
    public final String sETP = "eng type";
    public final String sMNF = "manufacturer";
    public final String sBYP = "b/p ratio";
    public final String sPRR = "pressure ratio";
    public final String sROP = "rated thrust (kN)";
    public final String sCMB = "combustor description";
    public final String sSNMAX = "sn max";
    public final String sTYR = "final test date";
    public final String sSSB = "data superseded";
    public final String sSSU = "superseded by uid";
    //
    // FOCA column headers
    private String[] FOCAMOD = { "_to", "_cl", "_app", "_taxi" };
    public final String sFOCA_UID = "code";
    public final String sFOCA_SRC = "Source";
    public final String sFOCA_MNF = "manufacturer_id";
    public final String sFOCA_ETP = "type_id";
    public final String sFOCA_ENG = "engine";
    public final String sFOCA_ROP = "power";
    public final String sFOCA_TYR = "mutationsdatum";
    //
    // FOI column headers
    public final String sFOI_MNF = "manufacturer";
    public final String sFOI_ENG = "engine id";
    public final String sFOI_ROP = "rated power [hp]";
    public int mod_offset = -1;
    //
    private HashMap<String, Integer> ht;
    public int iUID;   
    public int iENG;
    public int iMNF;
    public int iETP;
    public int iBYP;
    public int iPRR;
    public int iROP;
    public int iCMB;
    public int iTYR;
    public int iSNMAX;
    public int iSSB;
    public int iSSU;
    public int iPMM;
    
    public ColumnIndex() {
      iUID = -1;
      iENG = -1;
      iMNF = -1;
      iETP = -1;
      iBYP = -1;
      iPRR = -1;
      iROP = -1;
      iCMB = -1;
      iTYR = -1;
      iSSB = -1;
      iSSU = -1;
      iSNMAX = -1;
      iPMM = -1;
      ht = new HashMap<String, Integer>();
    }
    
    /**
     * Sets the column index for a given tracer and LTO mode
     * @param trc    the tracer name
     * @param mode   the mode name
     * @param d      the column index
     */
    public void set(String trc, String mode, int d) {    
      ht.put(trc.toLowerCase()+"-"+mode.toLowerCase(), d);
    }
    
    /**
     * Returns the column index for a given tracer and LTO mode
     * @param trc    the tracer name
     * @param mode   the mode name
     * @return the column index
     */
    public int get(String trc, String mode) {    
      Integer d = ht.get(trc.toLowerCase()+"-"+mode.toLowerCase());
      if (d == null)
        return -1;
      else
        return d;
    }
    
    /**
     * Sets the column indices for a EEDB in CSV format
     * @param cs   the CSV object
     * @param trca  the list of expected tracers
     * @throws Exception 
     */
    public void setCSVIndices(CST cs, String[] trca) throws Exception {
      setCSVIndices(cs, trca, false);
    }
    
    /**
     * Sets the column indices for a EEDB in CSV format
     * @param cs    the CSV object
     * @param trca  the list of expected tracers
     * @param check if true, a consistency check is performed
     * @throws Exception 
     */
    public void setCSVIndices(CST cs, String[] trca, boolean check) throws Exception {
      iUID = cs.getColumnIndex("UID");
      if (iUID < 0)
        iUID = 0;
      iENG = cs.getColumnIndex("ENG"); 
      iETP = cs.getColumnIndex("ETP"); 
      iBYP = cs.getColumnIndex("BYP"); 
      iPRR = cs.getColumnIndex("PRR"); 
      iROP = cs.getColumnIndex("ROP"); 
      iCMB = cs.getColumnIndex("CMB"); 
      iTYR = cs.getColumnIndex("TYR"); 
      iSNMAX = cs.getColumnIndex("SN-MAX");
      iPMM = cs.getColumnIndex("PMM");
      if (check) {
        if (iUID < 0 || iENG < 0 || iETP < 0 || iBYP < 0 || iPRR < 0 || 
          iROP < 0 || iCMB < 0 || iTYR < 0 || iSNMAX < 0) {
          String ss = String.format("(iUID=%d, iENG=%d, iETP=%d, iBYP=%d, iPRR=%d, iROP=%d, iCMB=%d, iTYR=%d, iSNMAX=%d)",
            iUID, iENG, iETP, iBYP, iPRR, iROP, iCMB, iTYR, iSNMAX);
          throw new Exception("basic columns not found! \n"+ss);
        }
      }
      for (String t : trca) {
        for (MOD.MODE m : MOD.MODE.values()) {
          int i = cs.getColumnIndex(t+"-"+m);
          if (i < 0) {
            if (check) throw new Exception("column not found for "+t+"-"+m);
          }
          else
            set(t, m.toString(), i);
        }
      }
    }
    
    /**
     * Sets the column indices for the original EEDB in XLS format
     * @param row  the row object with the column header
     * @param isheet  the XLS sheet index
     * @throws Exception 
     */
    public void setEEDBIndices(Row row, int isheet) throws Exception {
      for (int i=0; i<row.getLastCellNum(); i++) {
        XSSFCell c = (XSSFCell)row.getCell(i);
        if (c == null)
          continue;
        String s = c.getStringCellValue().trim();
        if (s.equalsIgnoreCase(sUID))
         iUID = i;
        else if (s.equalsIgnoreCase(sENG) )
          iENG = i;
        else if (s.equalsIgnoreCase(sMNF) )
          iMNF = i;
        else if (s.equalsIgnoreCase(sETP) )
          iETP = i;
        else if (s.equalsIgnoreCase(sBYP) )
          iBYP = i;
        else if (s.equalsIgnoreCase(sPRR) )
          iPRR = i;
        else if (s.equalsIgnoreCase(sROP) )
         iROP = i;
        else if (s.equalsIgnoreCase(sCMB) )
          iCMB = i;
        else if (s.toLowerCase().contains(sTYR) )
          iTYR = i;
        else if (s.toLowerCase().contains(sSNMAX) )
          iSNMAX = i;
        else if (s.toLowerCase().contains(sSSB) )
          iSSB = i;
        else if (s.toLowerCase().contains(sSSU) )
          iSSU = i;
      }
      if (iUID < 0 || iENG < 0 || iMNF < 0 || iETP < 0 || iBYP < 0 || iPRR < 0 || iROP < 0 || iCMB < 0 || iTYR < 0) {
        String ss = String.format("(iUID=%d, iENG=%d, iMNF=%d, iETP=%d, iBYP=%d, iPRR=%d, iROP=%d, iCMB=%d, iTYR=%d)",
          iUID, iENG, iMNF, iETP, iBYP, iPRR, iROP, iCMB, iTYR);
        throw new Exception("basic columns not found! \n"+ss);
      }
      //
      // tracers
      int it1 = (isheet == 0) ? 0 : 4;
      int it2 = (isheet == 0) ? 4 : EEDBTRCORG.length-1;
      for (int it=it1; it<=it2; it++) {
        for (int im=0; im<EEDBMOD.length; im++) {
          String smode_in = EEDBMOD[im];
          String smode_out = (im < EEDBMOD.length-1) ? MOD.MODE.values()[im].toString() : MOD.LTO; 
          boolean found = false;
          for (int i=0; i<row.getLastCellNum(); i++) {
            XSSFCell c = (XSSFCell)row.getCell(i);
            if (c == null)
              continue; 
            c.setCellType(CellType.STRING);
            String s = c.getStringCellValue().trim();
            s = s.toLowerCase();
            if (s.contains(EEDBTRCORG[it].toLowerCase()) && s.contains(smode_in.toLowerCase())) {
              set(EEDBTRCLSP[it], smode_out, i);
              found = true;
              break;
            }
          }
          if (!found && im < EEDBMOD.length-1) 
            throw new Exception("sheet "+isheet+": column not found for "+EEDBTRCORG[it]+" / "+EEDBMOD[im]);
        }
      }
    }
    
    /**
     * Sets the column indices for the FOCA data in XLS format
     * @param row  the row object with the column header
     * @throws Exception 
     */
    public void setFOCAIndices(Row row) throws Exception {
      for (int i=0; i<row.getLastCellNum(); i++) {
        XSSFCell c = (XSSFCell)row.getCell(i);
        if (c == null)
          continue;
        String s = c.getStringCellValue().trim();
        if (s.equalsIgnoreCase(sFOCA_UID))
          iUID = i;
        else if (s.equalsIgnoreCase(sFOCA_ENG) )
          iENG = i;
        else if (s.equalsIgnoreCase(sFOCA_MNF) )
          iMNF = i;
        else if (s.equalsIgnoreCase(sFOCA_ETP) )
          iETP = i; 
        else if (s.equalsIgnoreCase(sFOCA_ROP) )
          iROP = i; 
        else if (s.toLowerCase().contains(sFOCA_TYR) )
          iTYR = i;
      }
      if (iUID < 0 || iENG < 0 || iMNF < 0 || iETP < 0 || iROP < 0 || iTYR < 0) {
        String ss = String.format("(iUID=%d, iENG=%d, iMNF=%d, iETP=%d, iROP=%d, iTYR=%d)",
          iUID, iENG, iMNF, iETP, iROP, iTYR);
        throw new Exception("basic columns not found! \n"+ss);
      }
      //
      // tracers
      int it1 = 0;
      int it2 = FOCATRCORG.length-1;
      for (int it=it1; it<=it2; it++) {
        for (int im=0; im<FOCAMOD.length; im++) {
          String smode_in = FOCAMOD[im];
          String smode_out = MOD.MODE.values()[im].toString();
          boolean found = false;
          for (int i=0; i<row.getLastCellNum(); i++) {
            XSSFCell c = (XSSFCell)row.getCell(i);
            if (c == null)
              continue; 
            c.setCellType(CellType.STRING);
            String s = c.getStringCellValue().trim();
            s = s.toLowerCase();
            if (s.contains(FOCATRCORG[it].toLowerCase()) && s.contains(smode_in.toLowerCase())) {
              set(FOCATRCLSP[it], smode_out, i);
              found = true;
              break;
            }
          }
          if (!found) 
            throw new Exception("column not found for "+FOCATRCORG[it]+" / "+FOCAMOD[im]);
        }
      }
    }
    
    /**
     * Sets the column indices for the FOI data in XLS format
     * @param row   the row object with the column header
     * @throws Exception 
     */
    public void setFOIIndices(Row row) throws Exception {
      for (int i=0; i<row.getLastCellNum(); i++) {
        XSSFCell c = (XSSFCell)row.getCell(i);
        if (c == null) continue;
        String s = c.getStringCellValue().trim();
        if (s.equalsIgnoreCase(sFOI_ENG) )
          iENG = i;
        else if (s.equalsIgnoreCase(sFOI_MNF) )
          iMNF = i;
        else if (s.equalsIgnoreCase(sFOI_ROP) )
          iROP = i;  
      }
      if (iENG < 0 || iMNF < 0 || iROP < 0) {
        String ss = String.format("(iENG=%d, iMNF=%d, iROP=%d)", iENG, iMNF, iROP);
        throw new Exception("basic columns not found! \n"+ss);
      }
      //
      // identify tracer columns by fuel flow and assume TO / CO / AP / ID
      int[] fi = new int[8];
      int index = 0;
      for (int i=0; i<row.getLastCellNum(); i++) {
        XSSFCell c = (XSSFCell)row.getCell(i);
        if (c == null) continue;
        String s = c.getStringCellValue().trim();
        if (s.toLowerCase().startsWith("fuel"))
          fi[index++] = i;
      }
      if (index != 8)
        throw new Exception("unexpected number of fuel columns");
      int it1 = 0;
      int it2 = FOITRCORG.length-1;
      for (int it=it1; it<=it2; it++) {
        for (int imod=0; imod<2; imod++) {
          String trc = FOITRCLSP[it]+"/"+((imod == 0) ? "org" : "mod");
          for (int im=0; im<MOD.MODE.values().length; im++) {
            int i = 4*imod + im;
            int i1 = fi[i];
            int i2 = (i < fi.length-1) ? fi[i+1]-1 : row.getLastCellNum();
            for (i=i1; i<=i2; i++) {
              XSSFCell c = (XSSFCell)row.getCell(i);
              if (c.getStringCellValue().toLowerCase().contains(FOITRCORG[it].toLowerCase())) {
                set(trc, MOD.MODE.values()[im].toString(), i);
                break;
              }
            }
          }
        }
      }
    }
  } 
  // end class ColumnIndex
  
  /**
   * The EEDB object
   */
  public EEDB() {
    init();
  }
  
  private void init() {
    hp = new LinkedHashMap<String,PRM>();
    setPrm("FSC", Double.class, FSCDEFAULT);
    setPrm("S4TOS6", Double.class, S4TOS6DEFAULT);
    setPrm("H2O", Double.class, H2ODEFAULT);
    setPrm("EI_SOX", Double.class, EI_SOX);
    setPrm("EI_FB", Double.class, EI_FB);
    setPrm("EI_CO2", Double.class, EI_CO2);
    setPrm("BNZ_IN_HC", Double.class, BNZ_IN_HC);
    setPrm("ODOR_IN_HC", Double.class, ODOR_IN_HC);
    setPrm("PM25_IN_PM10", Double.class, PM25_IN_PM10);
    setPrm("USE_CERT_LTO", Integer.class, 1);
  }
  
  /**
   * Checks if the specified aircraft group name is valid
   * @param s the aircraft group name
   * @return 
   */
  public boolean isValidACG(String s) {
    if (s == null)
      return false;
    for (ACGT g : ACGT.values())
      if (g.toString().equalsIgnoreCase(s))
        return true;
    return false;
  }
  
  /**
   * Checks if the specified APU name is valid
   * @param s the APU name
   * @return 
   */
  public boolean isValidAPU(String s) {
    if (s == null)
      return false;
    for (APUT g : APUT.values())
      if (g.toString().equalsIgnoreCase(s))
        return true;
    return false;
  }
  
  /**
   * Checks if the specified UID is valid
   * @param s the UID
   * @return 
   */
  public boolean isValidUID(String s) {
    if (s == null)
      return false;
    return (get(s) != null);
  }
  
  /**
   * Checks if the specified number of engines is plausible
   * @param n the number of engines
   * @return 
   */
  public boolean isValidNEN(int n) {   
    return (n > 0 && n < 9);
  }
  
  private String getDateString(Row row, int i) {
    String s = "undef";
    if (i >= 0) {
      Cell c = row.getCell(i);
      if (c != null) {
        Date d = c.getDateCellValue();
        if (d != null) {
          SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
          s = sdf.format(d);
        }
      }
    }
    return s;
  }
  
  private String getString(Row row, int i) {
    if (i < 0)
      return null;
    Cell c = row.getCell(i);
    return (c == null || c.getCellType() == CellType.BLANK) ? "" : c.getStringCellValue(); 
  }
  
  private double getDouble(Row row, int i) {
    if (i < 0)
      return Double.NaN;
    Cell c = row.getCell(i);
    return (c == null || c.getCellType() == CellType.BLANK) ? Double.NaN : c.getNumericCellValue(); 
  }
  
  /**
   * Reads the EEDB in CSV format
   * @param fn        the file name including path information
   * @param section   the CSV section marker
   * @param trca      the tracer list
   * @param src       the data source information
   * @param check     if true, more checks are performed
   * @return { number of engines added, number of engines modified }
   * @throws Exception 
   */
  public int[] readCSV(String fn, String section, String[] trca, String src, boolean check) throws Exception {
    File f = new File(fn);
    if (!f.exists())
      throw new Exception("file "+fn+" not found!");
    BufferedReader br = new BufferedReader(new FileReader(fn));
    return readCSV(br, section, trca, src, check);
  }
  
  /**
   * 
   * Reads the EEDB in CSV format from a buffered reader
   * @param br        the buffered reader
   * @param section   the CSV section marker
   * @param trca      the tracer list
   * @param src       the data source information
   * @param check     if true, more checks are performed
   * @return { number of engines added, number of engines modified }
   * @throws Exception 
   */
  public int[] readCSV(BufferedReader br, String section, String[] trca, String src, boolean check) throws Exception {
    CST cs = new CST(br, section, SEPARATOR);
    if (cs.getRowCount() <= 0)
      return null;
    ColumnIndex ci = new ColumnIndex();
    ci.setCSVIndices(cs, trca, check);
    String s;
    double d;
    int nadded = 0;
    int nmodified = 0;
    for (int ir=0; ir<cs.getRowCount(); ir++) {
      String uid = cs.getString(ir, ci.iUID);
      if (uid.equalsIgnoreCase("UNIT")) {
        //
        // check units
        String su = cs.getString(ir, ci.iROP);
        if (check && !su.contains("kN") && !su.contains("hp"))
          throw new Exception("invalid unit for column ROP (expected kN/hp)");
        for (String t : trca) {
          su = Unit(DOMAIN.INPUT, GROUP.AC, t);
          for (int ic=0; ic<cs.getColumnCount(); ic++)
            if (cs.getColumnName(ic).toLowerCase().startsWith(t.toLowerCase()))
              if (!cs.getString(ir, ic).equalsIgnoreCase(su))
                throw new Exception("invalid unit for column "+cs.getColumnName(ic)+" (expected "+su+")"); 
        }
      }
      else { 
        //
        // read engine
        ENG eng = get(uid);
        if (eng == null) {
          eng = new ENG(uid);
          put(eng);
          nadded++;
        }
        else
          nmodified++;
        eng.ISRC = (section.toUpperCase().endsWith("ACG")) ? SOURCE.ACG.hashCode() : SOURCE.OTHER.hashCode();
        eng.SRC = src;
        if (eng.ISRC != SOURCE.ACG.hashCode() && isValidACG(eng.UID))
          throw new Exception("*** an engine UID may not be named like a default aircraft group: "+eng.UID); 
        s = cs.getString(ir, ci.iENG); if (s != null) eng.ENG = s;
        s = cs.getString(ir, ci.iETP); if (s != null) eng.ETP = s;
        d = cs.getDouble(ir, ci.iROP); if (!Double.isNaN(d)) eng.ROP = d;
        d = cs.getDouble(ir, ci.iBYP); if (!Double.isNaN(d)) eng.BYP = d;
        d = cs.getDouble(ir, ci.iPRR); if (!Double.isNaN(d)) eng.PRR = d;
        d = cs.getDouble(ir, ci.iSNMAX); if (!Double.isNaN(d)) eng.set(SMOKENUMBER, SNMAX, d); 
        s = cs.getString(ir, ci.iTYR); if (s != null) eng.FTD = s;
        s = cs.getString(ir, ci.iCMB); if (s != null) eng.CMB = s;
        eng.SSU = "";
        eng.FTD = "";
        s = cs.getString(ir, ci.iPMM); if (s != null) eng.PMM = s;
        if (eng.PMM == null)
          eng.PMM = UNDEF;
        for (String t : trca) {
          for (MOD.MODE m : MOD.MODE.values()) {
            d = cs.getDouble(ir, ci.get(t, m.toString()));
            if (!Double.isNaN(d) && d >= 0.)
              eng.set(t, m.toString(), d);
          }
        }
      }
    } 
    return new int[] { nadded, nmodified };
  }
  
  /**
   * Reads the Time-in-Mode table from a CSV file
   * @param fn       the file name including path
   * @param section  the section marker
   * @return number of aircraft groups read
   * @throws Exception 
   */
  public int readTIM(String fn, String section) throws Exception {
    return readTIM(fn, section, false); 
  }
  
  /**
   * Reads the Time-in-Mode table from a CSV file
   * @param fn       the file name including path
   * @param section  the section marker
   * @param cert_lto if true, the TIMS of the certification LTO are inserted
   * @return number of aircraft groups read
   * @throws Exception 
   */
  public int readTIM(String fn, String section, boolean cert_lto) throws Exception {
    File f = new File(fn);
    if (!f.exists())
      throw new Exception("file "+fn+" not found!");
    BufferedReader br = new BufferedReader(new FileReader(fn));
    CST cs = new CST(br, section, SEPARATOR);
    if (cs.getRowCount() <= 0)
      return 0;
    int nadded = 0;
    for (ACGT a : ACGT.values()) {
      int ir = cs.getRow(a.toString());
      if (ir < 0)
        continue;
      ACG acg = getACG(a.toString());
      if (acg == null) {
        acg = new ACG(a.toString());
        putACG(acg);      
      }
      String apu = cs.getString(ir, APU);
      if (apu != null)
        acg.APU = apu;
      for (MOD.MODE m : MOD.MODE.values()) {
        double d = cs.getDouble(ir, m.toString());
        if (cert_lto)
          d = MOD.ICAOTIM[m.ordinal()];
        if (!Double.isNaN(d) && d >= 0.)
          acg.setTIM(m.toString(), d);
      }
      for (APUM m : APUM.values()) {
        double d = cs.getDouble(ir, m.toString());
        if (!Double.isNaN(d) && d >= 0.)
          acg.setTIM(m.toString(), d);
      }
      nadded++;
    }
    return nadded;
  }

  /**
   * Reads the APU data from a CSV file
   * @param fn       the file name including path
   * @param section  the section marker
   * @param trca     the tracer list
   * @param src      data source information
   * @return number of APU types read
   * @throws Exception 
   */
  public int readAPU(String fn, String section, String[] trca, String src) throws Exception {
    double d;
    File f = new File(fn);
    if (!f.exists())
      throw new Exception("file "+fn+" not found!");
    BufferedReader br = new BufferedReader(new FileReader(fn));
    CST cs = new CST(br, section, SEPARATOR);
    if (cs.getRowCount() <= 0)
      return 0;
    int icn = cs.getColumnIndex("Name");
    int ict = cs.getColumnIndex("Tracer");
    int nadded = 0;
    for (APUT apu : APUT.values()) {
      boolean found = false;
      for (String t : trca) {
        int ir = 0;
        for (ir=0; ir<cs.getRowCount(); ir++) {
          if (cs.getString(ir, icn).equalsIgnoreCase(apu.toString()) &&
              cs.getString(ir, ict).equalsIgnoreCase(t))
            break;
        }
        if (ir >= cs.getRowCount())
          continue;
        found = true;
        String sunit = cs.getString(ir, UNIT);
        if (sunit == null)
          throw new Exception("APU unit specification required!");
        if (t.equalsIgnoreCase(FUELFLOW)) {
          if (!sunit.equalsIgnoreCase("kg/h"))
            throw new Exception("wrong APU unit for "+t);
        }
        else if (t.equalsIgnoreCase(NVPN)) {
          if (!sunit.equalsIgnoreCase("1/kg"))
            throw new Exception("wrong APU unit for "+t);
        }
        else if (!sunit.equalsIgnoreCase("g/kg"))
          throw new Exception("wrong APU unit for "+t);
        for (APUM m : APUM.values()) {
          d = cs.getDouble(ir, m.toString());
          if (!Double.isNaN(d) && d >= 0.)
            setAPU(apu.toString(), t, m, d);
        }
      }
      if (found)
        nadded++;
    }
    return nadded;
  }

  /**
   * Reads the original ICAO EEDB in XLS format
   * @param fn the file name including path information
   * @param src data source information
   * @return number of engines added
   * @throws Exception 
   */
  public int readEEDB(String fn, String src) throws Exception {
    File f = new File(fn);
    if (!f.exists())
      throw new Exception("file "+fn+" not found!");
    setPrm("EEDB", String.class, f.getName());
    src = src.replace(SEPARATOR, SEPREPLACE);
    InputStream istr = null;
    int nadded = 0;
    int irow = 0;
    int is = -1;
    int issue = 0;
    double warn_dev = 0.05;
    try {
      istr = new FileInputStream(f.getAbsolutePath());
      XSSFWorkbook wb = new XSSFWorkbook(istr); 
      XSSFSheet[] sheet = new XSSFSheet[2];
      //
      // get sheets
      String[] smarker = { "gaseous", "nvPM" };
      for (is=0; is<sheet.length; is++) {
        sheet[is] = null;
        for (int i=0; i<wb.getNumberOfSheets(); i++) {
          String name = wb.getSheetAt(i).getSheetName();
          if (name.toLowerCase().contains(smarker[is].toLowerCase())) {
            sheet[is] = wb.getSheetAt(i);
            break;
          }
        }
        if (sheet[is] == null)
          throw new Exception("sheet \""+smarker[is]+"\" not found!");
      }
      //
      // read
      nadded = 0;
      for (is=0; is<sheet.length; is++) {
        Iterator rows = sheet[is].rowIterator();
        Row row = (XSSFRow)rows.next();
        ColumnIndex ci = new ColumnIndex();
        ci.setEEDBIndices(row, is);
        irow = 1;
        while (rows.hasNext()) {
          irow++;
          row = (XSSFRow)rows.next();
          if (!row.cellIterator().hasNext())
            continue;
          String uid = getString(row, ci.iUID).trim();
          if (uid.length() == 0)                                  //-2023-07-04
            continue;
          ENG eng = get(uid);
          String seng = getString(row, ci.iENG);
          String smnf = getString(row, ci.iMNF);
          String setp = getString(row, ci.iETP);
          String scmb = getString(row, ci.iCMB);
          double dbyp = getDouble(row, ci.iBYP);
          if (Double.isNaN(dbyp))
            dbyp = 0;
          double drop = getDouble(row, ci.iROP);
          if (Double.isNaN(drop))
            drop = 0;
          double dprr = getDouble(row, ci.iPRR);
          if (Double.isNaN(dprr))
            dprr = 0;
          double dsnm = getDouble(row, ci.iSNMAX);
          if (Double.isNaN(dsnm))
            dsnm = -1;
          String styr = getDateString(row, ci.iTYR);
          if (styr.equalsIgnoreCase("undef"))
            styr = getDateString(row, ci.iTYR-1); 
          String sssb = getString(row, ci.iSSB);
          String sssu = getString(row, ci.iSSU);
          if (eng == null) {
            if (is > 0)
              throw new Exception("pm: engine "+uid+" not defined in sheet gas!");
            eng = new ENG(uid);
            eng.ISRC = SOURCE.ICAO.hashCode();
            eng.SRC = src;
            eng.ENG = seng;
            eng.MNF = smnf;
            eng.ETP = setp;
            eng.CMB = scmb;
            eng.BYP = dbyp;
            eng.ROP = drop;
            eng.PRR = dprr;
            eng.FTD = styr;
            eng.SSB = sssb.toLowerCase().contains("yes");
            eng.SSU = sssu;
            eng.PMM = UNDEF;
            eng.set(SMOKENUMBER, SNMAX, dsnm);
            put(eng);
            nadded++;
          }
          else {
            eng.ETP_PM = setp;
            eng.BYP_PM = dbyp;
            eng.PRR_PM = dprr;
            eng.ROP_PM = drop;
            if (!eng.ETP.equalsIgnoreCase(eng.ETP_PM))
              throw new Exception(String.format("%s deviation gas/pm, engine type: %s/%s: %s", eng.UID, eng.ETP, eng.ETP_PM, uid));
            if (Math.abs((eng.BYP-eng.BYP_PM)/eng.BYP) > 0.15 || 
                Math.abs((eng.ROP-eng.ROP_PM)/eng.ROP) > 0.15 || 
                Math.abs((eng.PRR-eng.PRR_PM)/eng.PRR) > 0.15)
              throw new Exception(String.format("%s strong deviations gas/pm, BYP=%1.4f/%1.4f, ROP=%1.4f/%1.4f, PRR=%1.4f/%1.4f", 
              eng.UID, eng.BYP, eng.BYP_PM, eng.ROP, eng.ROP_PM, eng.PRR, eng.PRR_PM));
            if (CHECK &&
               (!eng.ETP.equalsIgnoreCase(setp) || 
               Math.abs((eng.BYP-eng.BYP_PM)/eng.BYP) > warn_dev || 
               Math.abs((eng.ROP-eng.ROP_PM)/eng.ROP) > warn_dev || 
               Math.abs((eng.PRR-eng.PRR_PM)/eng.PRR) > warn_dev))
              System.out.printf("%04d %-16s, ETP=%5s/%5s, BYP=%8.4f/%8.4f, ROP=%8.4f/%8.4f, PRR=%8.4f/%8.4f\n", 
              ++issue, eng.UID, eng.ETP, eng.ETP_PM, eng.BYP, eng.BYP_PM, eng.ROP, eng.ROP_PM, eng.PRR, eng.PRR_PM);
          }
          for (int it=0; it<EEDBTRCLSP.length; it++) {
            String trc = EEDBTRCLSP[it];
            double fac = EEDB2LSP[it];
            for (MOD.MODE m : MOD.MODE.values()) {
              double d = getDouble(row, ci.get(trc, m.toString()));
              if (!Double.isNaN(d)) {
                d *= fac;
                double old = eng.get(trc, m.toString());   
                if (Double.isNaN(old)) {
                  if (is == 1 && !trc.equalsIgnoreCase(FUELFLOW)) {
                    //
                    // scale the EIs of the "pm" sheet with the ratio of
                    // fuel flows from the "pm" and the "gas" sheet so that
                    // the EIs from the "pm" sheet can be used in combination
                    // with the fuel flows from the gas sheet
                    double ffgas = eng.get(FUELFLOW, m.toString());
                    double ffpm = getDouble(row, ci.get(FUELFLOW, m.toString()));
                    if (Double.isNaN(ffgas) || Double.isNaN(ffpm) || Math.abs((ffgas-ffpm)/ffgas) > 0.15)
                      throw new Exception(String.format("%s: strong deviations gas/pm, fuel flow at %s: gas=%1.4f/%1.4f", 
                      eng.UID, m.toString(), ffgas, ffpm));
                    d *= (ffpm/ffgas);
                    eng.PMM = MEASURED;
                  }
                  eng.set(trc, m.toString(), d);
                }
                if (CHECK && !Double.isNaN(old) && d != old) {
                  double p = Math.abs(d-old)/(0.5*(d+old));
                  if (p > warn_dev)
                    System.out.printf("%04d %-16s, %2s/%2s: EI gas sheet: %12.5e, EI pm sheet: %12.5e (%3.1f%%)\n", 
                      ++issue, eng.UID, trc, m.toString(), old, d, 100*p);
                }
              }
            }
            double d = getDouble(row, ci.get(trc, LTO));
            if (!Double.isNaN(d)) {
              d *= fac;
              double old = eng.get(trc, LTO);
              if (Double.isNaN(old))
                eng.set(trc, LTO, d);
            }
          }
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace(System.out);
      System.out.printf("*** error in reading row "+irow+"\n");
      throw new Exception("can't extract data from sheet "+is);
    }
    if (CHECK) {
      //
      // check LTO emissions
      for (String st : EEDBTRCLSP) {
        ArrayList<String> va = getUIDs();
        for (String s : va) {
          ENG eng = get(s);
          double org = eng.get(st, LTO);
          if (!Double.isNaN(org)) {
            double d = eng.getLTOEmission(st);
            double p = Math.abs(d-org)/d;
            if (p > warn_dev && !eng.SSB)
              System.out.printf("%04d %-10s/%-5s, %s, LTO EEDB:%12.4f, calculated:%12.4f (%1.1f%%)\n", 
                ++issue, s, st, eng.FTD, org, d, 100*p);
          }
        }
      }
    }
    if (CHECK && issue > 0)
      System.out.printf("%d issues detected during check (treshold deviation: %1.2f)\n", issue, warn_dev);
    return nadded;
  }
  
  /**
   * Reads the original FOCA data in XLS format
   * @param fn the file name including path information
   * @param src data source information
   * @return number of engines added
   * @throws Exception 
   */
  public int readFOCA(String fn, String src) throws Exception {
    File f = new File(fn);
    if (!f.exists())
      throw new Exception("file "+fn+" not found!");
    setPrm("FOCA", String.class, f.getName());
    src = src.replace(SEPARATOR, SEPREPLACE);
    InputStream istr = null;
    int irow = 0;
    int nadded;
    try {
      istr = new FileInputStream(f.getAbsolutePath());
      XSSFWorkbook wb = new XSSFWorkbook(istr); 
      XSSFSheet sheet = null;
      for (int i=0; i<wb.getNumberOfSheets(); i++) {
        String name = wb.getSheetAt(i).getSheetName();
        if (name.toLowerCase().contains("data")) {
          sheet = wb.getSheetAt(i);
          break;
        }
      }
      if (sheet == null)
        throw new Exception("sheet containing \"data\" not found!");
      //
      // read
      Iterator rows = sheet.rowIterator();
      Row row = (XSSFRow)rows.next();
      ColumnIndex ci = new ColumnIndex();
      ci.setFOCAIndices(row);
      irow = 1;
      nadded = 0;
      while (rows.hasNext()) {
        irow++;
        row = (XSSFRow)rows.next();
        if (!row.cellIterator().hasNext())
          break;
        String uid = getString(row, ci.iUID);
        String seng = getString(row, ci.iENG);
        String smnf = getString(row, ci.iMNF);
        String setp = getString(row, ci.iETP);
        double drop = getDouble(row, ci.iROP); 
        String styr = getDateString(row, ci.iTYR); 
        ENG eng = new ENG(uid);
        eng.ISRC = SOURCE.FOCA.hashCode();
        eng.SRC = src;
        eng.ENG = seng;
        eng.MNF = smnf;
        eng.ETP = setp;
        eng.ROP = drop;
        eng.FTD = styr;
        eng.CMB = "";
        eng.SSU = "";
        eng.PMM = MEASURED;
        put(eng);
        nadded++;
        for (int it=0; it<FOCATRCLSP.length; it++) {
          String trc = FOCATRCLSP[it];
          double fac = FOCA2LSP[it];
          for (MOD.MODE m : MOD.MODE.values()) {
            double d = getDouble(row, ci.get(trc, m.toString()));
            if (!Double.isNaN(d))
              d *= fac;
            eng.set(trc, m.toString(), d);
          }
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace(System.out);
      System.out.printf("*** error in reading row "+irow+"\n");
      throw new Exception("can't extract data");
    }
    finally {
      try { istr.close(); } catch (Exception ee) {};
    }
    return nadded;
  }
  
  /**
   * Reads the original FOI data in XLS format
   * @param fn the file name including path information
   * @param src data source information
   * @return number of engines added
   * @throws Exception 
   */
  public int readFOI(String fn, String src) throws Exception {
    File f = new File(fn);
    if (!f.exists())
      throw new Exception("file "+fn+" not found!");
    setPrm("FOI", String.class, f.getName());
    src = src.replace(SEPARATOR, SEPREPLACE);
    int irow = 1;
    int nadded;
    InputStream  istr = null;
    try {
      istr = new FileInputStream(f.getAbsolutePath());
      XSSFWorkbook wb = new XSSFWorkbook(istr); 
      XSSFSheet sheet = null;
      for (int i=0; i<wb.getNumberOfSheets(); i++) {
        String name = wb.getSheetAt(i).getSheetName();
        if (name.toLowerCase().contains("tpeng")) {
          sheet = wb.getSheetAt(i);
          break;
        }
      }
      if (sheet == null)
        throw new Exception("sheet containing \"TPeng\" not found!");
      //
      // read
      Iterator rows = sheet.rowIterator();
      Row row;
      while (null != (row = (XSSFRow)rows.next())) {
          XSSFCell c = (XSSFCell)row.getCell(0);
          if (c == null) continue;
          String s = c.getStringCellValue().trim();
          if (s.equalsIgnoreCase("manufacturer"))
            break;
      }
      if (row == null) 
        throw new Exception("can't dind starting line");
      ColumnIndex ci = new ColumnIndex();
      ci.setFOIIndices(row);
      irow = 1;
      nadded = 0;
      int index = 1;
      while (rows.hasNext()) {
        irow++;
        row = (XSSFRow)rows.next();
        if (!row.cellIterator().hasNext())
          break;
        String seng = getString(row, ci.iENG);
        String smnf = getString(row, ci.iMNF);
        double drop = getDouble(row, ci.iROP); 
        ENG eng = new ENG(String.format("FOI%03d", index++));
        eng.ISRC = SOURCE.FOI.hashCode();
        eng.SRC = src;
        eng.ENG = seng;
        eng.MNF = smnf;
        eng.ETP = "";
        eng.ROP = drop;
        eng.FTD = "";
        eng.CMB = "";
        eng.SSU = "";
        eng.PMM = UNDEF;
        put(eng);
        nadded++;
        for (int it=0; it<FOITRCLSP.length; it++) {
          String trc = FOITRCLSP[it];
          double fac = FOI2LSP[it];
          for (MOD.MODE m : MOD.MODE.values()) {
            double dorg = getDouble(row, ci.get(trc+"/org", m.toString()));
            double dmod = getDouble(row, ci.get(trc+"/mod", m.toString()));
            double d;
            if (Double.isNaN(dorg))
              ; //System.out.printf("### %s/%s/%s: dorg=null\n", seng, trc, sm);
            if (!Double.isNaN(dorg) && !Double.isNaN(dmod))
             ; // if (Math.abs(dorg-dmod)/dmod > 0.5) System.out.printf("### %s/%s/%s: dorg=%e, dmod=%e\n", seng, trc, sm, dorg, dmod);
            if (!Double.isNaN(dmod)) {
              d = dmod;
              // some entries of the modified FOI table are negative! -2022-07-14
              if (d < 0)
                d = 0.;
            }
            else
              d = dorg;
            if (!Double.isNaN(d))
              d *= fac;
            eng.set(trc, m.toString(), d);
          }
        }
      }
    }
    catch (Exception e) {
      e.printStackTrace(System.out);
      System.out.printf("*** error in reading row "+irow+"\n");
      throw new Exception("can't extract data");
    }
    finally {
      try { istr.close(); } catch (Exception ee) {}
    }
    return nadded;
  }
  
  /**
   * Returns all UIDs (without surrogates)
   * @return 
   */
  public ArrayList<String> getUIDs() {
    return getUIDs(false, false);
  }
  
  /**
   * Returns all UIDs with measured PM EIs (without surrogates)
   * @return 
   */
  public ArrayList<String> getUIDsPMM() {
    return getUIDs(false, true);
  }
  
  /**
   * Returns all UIDs
   * @param withsurrogates
   * @param measuredonly
   * @return 
   */
  public ArrayList<String> getUIDs(boolean withsurrogates, boolean measuredonly) {
    if (hm != null) {
      ArrayList<String> va = new ArrayList<String>();
      for (String s : hm.keySet()) {
        ENG e = hm.get(s);
        if (measuredonly && (e.PMM == null || !e.PMM.equals(MEASURED)))
          continue;
        va.add(e.UID);
        if (withsurrogates && e.SRG != null)
          for (String ss : e.SRG)
            va.add(ss);
      }
      return va;
    }
    else
      return null;
  }
  
  /**
   * Sets an APU emission
   * @param apu  the APU name
   * @param trc  the tracer name
   * @param mode the APU mode
   * @param d the emission value
   */
  public void setAPU(String apu, String trc, APUM mode, double d) {
    if (hapu == null)
      hapu = new LinkedHashMap<String, Double>();
    hapu.put(apu+"-"+mode+"-"+trc.toUpperCase(), d);
  }

  /**
   * Gets an APU emission
   * @param apu  the APU name
   * @param trc  the tracer name
   * @param mode the APU mode
   * @return the emission value
   */
  public double getAPU(String apu, String trc, APUM mode) {
    if (hapu != null) {
      Double D = hapu.get(apu+"-"+mode+"-"+trc.toUpperCase());
      return (D == null) ? Double.NaN : D;
    }
    else
      return Double.NaN;
  }

  /**
   * Gets an engine
   * @param uid the UID
   * @return the engine object
   */
  public ENG get(String uid) {
    if (uid == null || hm == null)
      return null;
    ENG e = hm.get(uid.toUpperCase());
    if (e == null) {
      // try surrogates
      for (String s : hm.keySet()) {
        e = hm.get(s);
        if (e.SRG != null)
          for (String ss : e.SRG)
            if (ss.equalsIgnoreCase(uid))
              return e;  
      }
      e = null;
    }
    return e;
  }
  
  /**
   * Gets an aircraft group
   * @param acg the group name
   * @return the group object
   */
  public ACG getACG(String acg) {
   if (acg != null && hg != null)
      return hg.get(acg.toUpperCase());
   else
     return null;
  }
  
  /**
   * Removes an engine
   * @param uid the UID
   */
  public void remove(String uid) {
   if (uid != null && hm != null)
     hm.remove(uid.toUpperCase());
  }
  
  /**
   * Clears the engine list
   */
  public void clear() {
    if (hm != null)
      hm.clear();
  }

  /**
   * Adds an engine
   * @param eng the engine
   */
  public void put(ENG eng) {
    if (hm == null)
      hm = new LinkedHashMap<String,ENG>();
    if (eng == null || eng.UID == null)
      return;
    if (hm.get(eng.UID.toUpperCase()) != null) {
      hm.remove(eng.UID.toUpperCase()); // overwrite existing engines
    }
    hm.put(eng.UID.toUpperCase(), eng);  
  }
  
  /**
   * Adds an aircraft group
   * @param acg the group
   */
  public void putACG(ACG acg) {
    if (hg == null)
      hg = new LinkedHashMap<String,ACG>();
    if (acg == null || acg.name == null)
      return;
    if (hg.get(acg.name.toUpperCase()) != null) {
      hg.remove(acg.name.toUpperCase());
    }
    hg.put(acg.name.toUpperCase(), acg);  
  }
  
  /**
   * Unsets a parameter
   * @param name the name of teh parameter
   */
  public void unsetPrm(String name) {
    if (name == null)
      return;
    PRM p = hp.get(name.toUpperCase());
    if (p != null)
      hp.remove(name.toUpperCase());
  }
  
  /**
   * Sets a parameter
   * @param name the name of the parameter
   * @param cl the class of the parameter value
   * @param v the parameter value
   */
  public void setPrm(String name, Class cl, Object v) {
    if (name == null || v == null)
      return;
    PRM p = hp.get(name.toUpperCase());
    if (p == null)
      hp.put(name.toUpperCase(), new PRM(name, cl, v));
    else {
      if (cl == Double.class && !Double.isNaN((Double)v)) {
        p.type = cl;
        p.value = v;
      }
      else if (cl == Integer.class && ((Integer)v) != Integer.MIN_VALUE) {
        p.type = cl;
        p.value = v;
      }
      else if (cl == String.class && ((String)v) != null) {
        p.type = cl;
        p.value = v;
      } 
    }
  }
  
  /**
   * Returns the parameter value if string
   * @param name the name of the parameter
   * @return 
   */
  public String getPrmString(String name) {
    PRM p = hp.get(name.toUpperCase());
    return (p == null) ? null : (String)p.value;
  }
  
  /**
   * Returns the parameter value if double
   * @param name the name of the parameter
   * @return 
   */
  public double getPrmDouble(String name) {
    PRM p = hp.get(name.toUpperCase());
    return (p == null) ? Double.NaN : (Double)p.value;
  }
  
  /**
   * Returns the parameter value if integer
   * @param name the name of the parameter
   * @return 
   */
  public int getPrmInt(String name) {
    PRM p = hp.get(name.toUpperCase());
    return (p == null) ? Integer.MIN_VALUE : (Integer)p.value;
  }
  
  /**
   * Checks if a modelled PM EI can be inserted
   * @param e the engine
   * @return true if a modelled PM EI can be inserted
   */
  boolean canInsertPM(ENG e) {
    return (e.ISRC == EEDB.SOURCE.ICAO.hashCode() ||
            e.ISRC == EEDB.SOURCE.FOI.hashCode() ||
            e.ISRC == EEDB.SOURCE.OTHER.hashCode());
  }
  
  /**
   * Checks if other derived EIs can be inserted
   * @param e the engine
   * @return true if other derived EIs can be inserted
   */
  boolean canInsertOther(ENG e) {
    return (e.ISRC == EEDB.SOURCE.ICAO.hashCode() ||
            e.ISRC == EEDB.SOURCE.FOI.hashCode() ||
            e.ISRC == EEDB.SOURCE.FOCA.hashCode() ||
            e.ISRC == EEDB.SOURCE.OTHER.hashCode() ||
            e.ISRC == EEDB.SOURCE.ACG.hashCode());
  }
  
  /**
   * Inserts all PM methods into all engines with loss correctionm
   * @throws Exception 
   */
  public void insertPM() throws Exception {
    insertPM(null, false);
  }
  
  /**
   * PM EIs according to different methodologies are stored in separate
   * fields of the ENG map; fields NVPM and NVPN are not changed.
   * @param usr   user defined parameter values for method FOA4USR
   * @param uncorrected true if no loss correction is applied
   * @throws Exception 
   */
  public void insertPM(double[] usr, boolean uncorrected) throws Exception {
    ArrayList<String> va = getUIDs();
    PM.METHOD[] ma = PM.METHOD.values();
    for (PM.METHOD m : ma) {
      PM pm = new PM(m, uncorrected);
      double fsc = getPrmDouble("FSC");
      double s4tos6 = getPrmDouble("S4TOS6");
      double h2o = getPrmDouble("H2O");
      if (!Double.isNaN(fsc))
        pm.setFsc(fsc);
      if (!Double.isNaN(s4tos6))
        pm.setS4toS6(s4tos6);
      if (!Double.isNaN(h2o))
        pm.setH2O(h2o);
      if (m == PM.METHOD.FOA4USR && usr != null && usr.length == 8) {
        pm.FOA4USR_gmd[0] = usr[0];
        pm.FOA4USR_gmd[1] = usr[1];
        pm.FOA4USR_gmd[2] = usr[2];
        pm.FOA4USR_gmd[3] = usr[3];
        pm.FOA4USR_gsd[0] = usr[4];
        pm.FOA4USR_gsd[1] = usr[5];
        pm.FOA4USR_gsd[2] = usr[6];
        pm.FOA4USR_gsd[3] = usr[7];
      }
      setPrm(m.toString(), String.class, pm.getInfo(m));
      for (String uid : va) {
        ENG e = get(uid);
        if (!canInsertPM(e))
          continue;
        pm.setEI(e);
      }
    }
  }
  
  /**
   * Copy the values of the specified method to fields NVPM and NVPN.
   * @param method  the method
   * @param uncorrected true if no loss correction is applied
   * @param overwrite true if existing values (measured) are overwritten
   * @throws Exception 
   */
  public void copyMethod(PM.METHOD method, boolean uncorrected, boolean overwrite) 
  throws Exception {
    setPrm("PM_Method", String.class, method.toString());
    ArrayList<String> va = getUIDs();
    String uc = (uncorrected) ? EEDB.X_UNCORRECTED : "";
    String[] pma = { NVPM+uc, NVPN+uc };
    for (String uid : va) {
      ENG e = get(uid);
      if (!canInsertPM(e)) 
        continue;
      if (e.PMM.equals(UNDEF) || overwrite)
        e.PMM = method.toString();
      double dnew, dold;
      for (MOD.MODE m : MOD.MODE.values()) {
        for (String pm : pma) {
          dnew = e.get(pm+"_"+method, m.toString());
          dold = e.get(pm, m.toString());
          if (Double.isNaN(dold) || dold < 0. || overwrite) 
            e.set(pm, m.toString(), dnew);
        }
      }
    }
  }
  
  /**
   * Inserts other derived EIs to all engines
   * @throws Exception 
   */
  public void insertOther() throws Exception {
    //
    // engines
    ArrayList<String> va = getUIDs();
    double d;
    //
    // set SOx emission index for engines and APU
    double fsc = getPrmDouble("FSC");
    double ei_sox = Double.NaN;
    if (!Double.isNaN(fsc) && fsc >= 0)
      ei_sox = fsc*1000*(64./32.);
    if (Double.isNaN(ei_sox))
      ei_sox = getPrmDouble("EI_SOX");
    //
    for (String uid : va) {
      ENG e = get(uid);
      if (!canInsertOther(e)) 
        continue;
      for (MOD.MODE m : MOD.MODE.values()) {
        e.set(SOX, m.toString(), ei_sox);
        e.set(CO2, m.toString(), getPrmDouble("EI_CO2"));
        e.set(FUELBURN, m.toString(), getPrmDouble("EI_FB"));
        d = e.get(HC, m.toString());
        if (!Double.isNaN(d) && d >= 0.) {
          e.set(BNZ, m.toString(), d * getPrmDouble("BNZ_IN_HC"));
          e.set(ODOR, m.toString(), d * getPrmDouble("ODOR_IN_HC"));
        }
        d = e.get(PM10, m.toString());
        if (!Double.isNaN(d) && d >= 0.)
          e.set(PM25, m.toString(), d * getPrmDouble("PM25_IN_PM10"));
      }
    }
    //
    // APU
    for (APUT apu : APUT.values()) {
      for (APUM m : APUM.values()) {
        setAPU(apu.toString(), SOX, m, ei_sox);
        setAPU(apu.toString(), CO2, m, getPrmDouble("EI_CO2"));
        setAPU(apu.toString(), FUELBURN, m, getPrmDouble("EI_FB"));
        d = getAPU(apu.toString(), HC, m);
        if (!Double.isNaN(d) && d >= 0.) {
          setAPU(apu.toString(), BNZ, m, d * getPrmDouble("BNZ_IN_HC"));
          setAPU(apu.toString(), ODOR, m, d * getPrmDouble("ODOR_IN_HC"));
        }
        d = getAPU(apu.toString(), PM10, m);
        if (!Double.isNaN(d) && d >= 0.)
          setAPU(apu.toString(), PM25, m, d * getPrmDouble("PM25_IN_PM10"));
      }
    }
  }
  
  /** 
   * Checks if the tracer is of type particulate number
   * @param name the tracer name
   * @return true if the name ends with pn
   */
  public static boolean IsNumberTracer(String name) {
    return (name != null && name.toLowerCase().contains("pn"));
  }
  
  /**
   * Checks if the tracer is of type odor
   * @param name the tracer name
   * @return true if the name contains odor
   */
  public static boolean IsOdorTracer(String name) {
    return (name != null && name.toLowerCase().contains("odor"));
  }
  
  /**
   * Returns the required unit of an emission value
   * @param domain the domain 
   * @param group the group
   * @param tracer the tracer name
   * @return 
   */
  public static String Unit(DOMAIN domain, GROUP group, String tracer) {
    String unit = null;
    if (domain == DOMAIN.INPUT) {
      if (group == GROUP.AC || group == GROUP.APU) {
        if (tracer.equalsIgnoreCase(FUELFLOW))
          unit = (group == GROUP.AC) ? "kg/s" : "kg/h";
        else if (IsOdorTracer(tracer))
          unit = "OU/kg";
        else if (IsNumberTracer(tracer))
          unit = "1/kg";
        else if (tracer.equalsIgnoreCase("SN"))
          unit = "1";
        else if (tracer.toLowerCase().startsWith("gmd"))
          unit = "nm";
        else if (tracer.toLowerCase().startsWith("cee"))
          unit = "g/m3";
        else
          unit = "g/kg";
      }
    }
    else if (domain == DOMAIN.EMIS) {
      if (IsOdorTracer(tracer))
        unit = "MOU";
      else if (IsNumberTracer(tracer))
        unit = "1";
      else
        unit = "Mg";
    }
    return unit;
  }
  
  private String makeOut(String s) {
    if (s == null)
      return "";
    s = s.trim();
    s = s.replaceAll(OUTEXCLUDE, "#");
    s = s.replace(SEPARATOR, SEPREPLACE);
    if (s.length() > 0)
      return "\""+s+"\"";
    else
      return "";
  }
  
  /**
   * Writes the time-in-mode table
   * @param fn the file name including path information
   * @throws Exception 
   */
  public void writeTIM(String fn) throws Exception {
    File f = new File(fn);
    BufferedWriter bw;
    bw = new BufferedWriter(new FileWriter(f));
    writeTIM(bw);
    bw.close();
  }
  
  /**
   * Writes the time-in-mode table
   * @param bw the buffered writer
   * @throws Exception 
   */
  public void writeTIM(BufferedWriter bw) throws Exception {
    StringBuilder sb = new StringBuilder();
    bw.newLine();
    bw.write("[TABLE.LTO.SECONDS]"); bw.newLine();
    sb.append(String.format("%-12s", "ACG"));
    for (MOD.MODE m : MOD.MODE.values())
      sb.append(String.format("; %8s", m));
    sb.append(String.format("; %-6s", APU));
    for (APUM m : APUM.values())
      sb.append(String.format("; %8s", m));
    bw.write(sb.toString()); bw.newLine();
    for (ACGT a : ACGT.values()) {
      ACG acg = getACG(a.toString());
      if (acg == null)
        continue;
      sb.setLength(0);
      sb.append(String.format("%-12s", a));
      for (MOD.MODE m : MOD.MODE.values())
        sb.append(String.format("; %8.1f", acg.getTIM(m.toString())));
      sb.append(String.format("; %-6s", acg.APU));
      for (APUM m : APUM.values())
        sb.append(String.format("; %8.1f", acg.getTIM(m.toString())));
      bw.write(sb.toString()); bw.newLine();
    }
  }
  
   /**
   * Writes the APU table
   * @param fn the file name including path
   * @param trca the tracer list
   * @throws Exception 
   */
  public void writeAPU(String fn, String[] trca) throws Exception {
    File f = new File(fn);
    BufferedWriter bw;
    bw = new BufferedWriter(new FileWriter(f));
    writeAPU(bw, trca);
    bw.close();
  }
  
  /**
   * Writes the APU table
   * @param bw the buffered writer
   * @param trca the tracer list
   * @throws Exception 
   */
  public void writeAPU(BufferedWriter bw, String[] trca) throws Exception {
    StringBuilder sb = new StringBuilder();
    double d;
    bw.newLine();
    bw.write("[TABLE.EEDB.APU]"); bw.newLine();
    sb.append(String.format("%-12s; %-12s; %-6s", "Name", "Tracer", "Unit"));
    for (APUM m : APUM.values())
      sb.append(String.format("; %12s", m));
    bw.write(sb.toString()); bw.newLine();
    for (String t : trca) {
      String unit = "g/kg";
      if (t.equalsIgnoreCase(SMOKENUMBER))
       continue;
      if (t.equalsIgnoreCase(FUELFLOW))
        unit = "kg/h";
      else if (IsNumberTracer(t))
        unit = "1/kg";
      else if (IsOdorTracer(t))
        unit = "OU/kg";
      for (APUT apu : APUT.values()) {
        sb.setLength(0);
        sb.append(String.format("%-12s; %-12s; %-6s", apu, t, unit));
        for (APUM m : APUM.values()) {
          d = getAPU(apu.toString(), t, m);
          if (Double.isNaN(d) || d < 0.)
            d = -999.;
          sb.append(String.format("; %12.5e", d));
        }
        bw.write(sb.toString()); bw.newLine();
      }  
    }
  }
  
  /**
   * Writes the EEDB
   * @param fn the file name including path
   * @param trca the tracer list
   * @throws Exception 
   */
  public void writeEEDB(String fn, String[] trca) throws Exception {
    writeEEDB(fn, SOURCE.UNDEF, trca);
  }
  
  /**
   * Writes the EEDB
   * @param bw the buffered writer
   * @param trca the tracer list
   * @throws Exception 
   */
  public void writeEEDB(BufferedWriter bw, String[] trca) throws Exception {
    write(bw, SOURCE.UNDEF, trca);
  }
  
  /**
   * Writes the EEDB with engines of type <source>
   * @param fn the file name including path
   * @param source the source
   * @param trca the tracer list
   * @throws Exception 
   */
  public void writeEEDB(String fn, SOURCE source, String[] trca) throws Exception {
    File f = new File(fn);
    BufferedWriter bw;
    bw = new BufferedWriter(new FileWriter(f));
    write(bw, source, trca);
    bw.close();
    System.out.printf("%s written.\n", f.getAbsolutePath());
  }
  
  /**
   * Writes the EEDB with engines of type <source>
   * @param bw the buffered writer
   * @param source the source
   * @param trca the tracer list
   * @throws Exception 
   */
  public void writeEEDB(BufferedWriter bw, SOURCE source, String[] trca) throws Exception {
    write(bw, source, trca);
  }
  

  private void write(BufferedWriter bw, SOURCE source, String[] trca) throws Exception {
    long l = Calendar.getInstance().getTimeInMillis();
    TimeZone tz = TimeZone.getTimeZone("GMT+01:00");
    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'"+"T"+"'HH:mm:ssZ");
    sdf.setTimeZone(tz);
    String dstr = sdf.format(new Date(l));
    bw.newLine();
    bw.write("[PARAMETER.EEDB]"); bw.newLine();
    bw.write("Creator       ; de.janicke.airpport.eedb.EEDB "+VERSION); bw.newLine();
    bw.write("Date          ; "+dstr); bw.newLine();
    bw.write("EEDB          ; "+getPrmString("EEDB")); bw.newLine();
    bw.write("FOCA          ; "+getPrmString("FOCA")); bw.newLine();
    bw.write("FOI           ; "+getPrmString("FOI")); bw.newLine();
    bw.write("UID           ; engine UID"); bw.newLine();
    bw.write("ENG           ; engine name"); bw.newLine();
    bw.write("SRC           ; data source"); bw.newLine();
    bw.write("CMB           ; combustor description"); bw.newLine();
    bw.write("ETP           ; engine type"); bw.newLine();
    bw.write("BYP           ; bypass ratio"); bw.newLine();
    bw.write("PRR           ; pressure ratio"); bw.newLine();
    bw.write("ROP           ; rated output"); bw.newLine();
    bw.write("TYR           ; date of data set"); bw.newLine();
    bw.write("<name>-<mode> ; fuel flows and EIs"); bw.newLine();
    bw.write("INVALID       ; -999.0"); bw.newLine();
    //
    bw.write("PM_Method     ; "+getPrmString("PM_Method")); bw.newLine();
    for (PM.METHOD m : PM.METHOD.values()) {
      bw.write(String.format("%-10s    ; %s", m, getPrmString(m.toString()))); bw.newLine();
    }
    bw.newLine();
    //
    // tracer list
    ArrayList<String> va = getUIDs();
    double d = get(va.get(0)).get(NVPM, MOD.MODE.TO.toString());
    ArrayList<String> vtrc = new ArrayList<String>();
    vtrc.addAll(Arrays.asList(trca));
    vtrc.remove(NO2);
    for (SOURCE src : SOURCE.values()) {
      if (src == SOURCE.UNDEF  || (source != SOURCE.UNDEF && src != source))
        continue;
      bw.newLine();
      bw.write("[TABLE.EEDB."+src+"]");
      bw.newLine();
      StringBuilder sb = new StringBuilder();
      //
      // header
      sb.append(String.format("%-12s; %-40s; %-10s; %-12s; %-40s; %-5s; %7s; %7s; %7s; %12s; %12s",
        "UID", "ENG", "SRC", "SSE", "CMB", "ETP", "BYP", "PRR", "ROP", "TYR", "PMM"));
      for (String trc : vtrc) {
        for (MOD.MODE m : MOD.MODE.values())
          sb.append(String.format("; %18s", trc + "-" + m));
        if (trc.equalsIgnoreCase(SMOKENUMBER))
          sb.append(String.format("; %18s", trc + "-MAX"));
      }
      bw.write(sb.toString());
      bw.newLine();
      //
      // units
      sb.setLength(0);
      sb.append(String.format("%-12s; %-40s; %-10s; %-12s; %-40s; %-5s; %7s; %7s; %7s; %12s; %12s",
        "Unit", "", "", "", "", "", "1", "1", "kN", "", ""));
      for (String trc : vtrc) {
        for (MOD.MODE m : MOD.MODE.values())
          sb.append(String.format("; %18s", Unit(DOMAIN.INPUT, GROUP.AC, trc)));
        if (trc.equalsIgnoreCase(SMOKENUMBER))
          sb.append(String.format("; %18s", "1"));
      }   
      bw.write(sb.toString());
      bw.newLine();
      //
      for (int block=0; block<2; block++) {  // 0: engines, 1: surrogates
        for (String uid : va) {
          ENG eng = get(uid);
          //System.out.printf("### block=%d, ISRC=%d, uid=%s\n", block, eng.ISRC, eng.UID);
          if (eng.ISRC != src.hashCode())
            continue;
          if (block > 0 && eng.SRG == null)
            continue;
          int is1=0, is2=0;
          if (block > 0)
            is2 = eng.SRG.length-1;
          for (int is=is1; is<=is2; is++) {
            sb.setLength(0);
            sb.append(String.format("%-12s; %-40s; %-10s; %-12s; %-40s; %-5s; %7.2f; %7.2f; %7.2f; %12s; %12s",
              (block == 0) ? eng.UID : eng.SRG[is].toUpperCase(), makeOut(eng.ENG), makeOut(eng.SRC), 
              (block == 0) ? eng.SSU : eng.UID, 
              makeOut(eng.CMB), makeOut(eng.ETP), 
              Double.isNaN(eng.BYP) ? -999.0: eng.BYP, 
              Double.isNaN(eng.PRR) ? -999.0 : eng.PRR, 
              (src != SOURCE.ICAO || Double.isNaN(eng.ROP)) ? -999.0 : eng.ROP, 
              makeOut(eng.FTD),
              makeOut(eng.PMM)));
            for (String trc : vtrc) {
              for (MOD.MODE m : MOD.MODE.values()) {
                d = eng.get(trc, m.toString());
                if (Double.isNaN(d) || d < 0.) 
                  d = -999.0;
                if (eng.UID.equalsIgnoreCase("1ZM001") && m == MOD.MODE.ID)
                  d = 0.;
                if (trc.equalsIgnoreCase(FUELFLOW) && d < 0)
                  throw new Exception("negative FF for "+eng.UID+"-"+m+", should be corrected!");
                sb.append(String.format("; %18.5e", d));
              }
              if (trc.equalsIgnoreCase(SMOKENUMBER)) {
                double snmax = eng.get(trc, SNMAX);
                sb.append(String.format("; %18.2f", (Double.isNaN(snmax)) ? -999.0 : snmax));
              }
            }
            bw.write(sb.toString());
            bw.newLine();
          }
        }
      }
    }
  }

  private void testCalvert() throws Exception {
    CHECK = true;
    File fin = new File("D:/p/ec-raptor/work/eedb", "edb-emissions-databank v28b (web).xlsx");
    readEEDB(fin.getAbsolutePath(), "28B");
    ENG eng = get("1PW010");
    PM pm = new PM(PM.METHOD.FOA3N);
    double snmax = eng.get(EEDB.SMOKENUMBER, EEDB.SNMAX);
    double[] sn = new double[MOD.MODE.values().length];
    for (MOD.MODE m : MOD.MODE.values()) {
      sn[m.ordinal()] = eng.get(EEDB.SMOKENUMBER, m.toString());
    }
    boolean mod = pm.reviewSN(sn, snmax, eng.UID, eng.ENG, eng.CMB, eng.MNF);
    for (MOD.MODE m : MOD.MODE.values())
      eng.set(EEDB.SMOKENUMBER, m.toString(), sn[m.ordinal()]);
    System.out.printf("%s\n", eng.toString()); 
  }
 
 
  public static void main(String[] args) {
    Locale.setDefault(Locale.ENGLISH);
    try {
      EEDB eedb = new EEDB();
      eedb.testCalvert();
    }
    catch (Exception e) {
      e.printStackTrace(System.out);
    }
    System.exit(0);
  }
 
}
