////////////////////////////////////////////////////////////////////////////////
//
// history:
//
// 2020-03-14  uj  1.0.0  setup
//
////////////////////////////////////////////////////////////////////////////////

/**
 * (c) Janicke Consulting, Germany 2022
 * 
 * Class to handle engine data.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

package de.janicke.airport.eedb;

import java.util.HashMap;

public class ENG {
  
    public static final String VERSION = "1.0.0";   
    private HashMap<String, Double> ht;
    public String UID;
    public String ENG;
    public String MNF;
    public String ETP; 
    public String ETP_PM; 
    public String CMB; 
    public double BYP;
    public double BYP_PM;
    public double PRR;
    public double PRR_PM;
    public double ROP;
    public double ROP_PM;
    public String FTD;
    public int ISRC;
    public String SRC;
    public boolean SSB;
    public String SSU;
    public String[] SRG;
    public String PMM;
    
    /**
     * Creates a new engine
     * @param uid engine UID
     */
    public ENG(String uid) {
      UID = uid;
      ht = new HashMap<String, Double>();
    }
    
    /**
     * DAC engine or not
     * @return true if the combustor specification contains the entry DAC
     */
    public boolean isDAC() {
      if (CMB != null && CMB.toLowerCase().contains("dac"))
        return true;
      else 
        return false;
    }
    
    /**
     * mixed turbofan engine or not
     * @return true if the engine type description contains the entry MTF
     */
    public boolean isMTF() {
      if (ETP != null && ETP.toLowerCase().contains("mtf"))
        return true;
      else 
        return false;
    }
    
    /** 
     * Sets an emission value
     * @param trc  tracer name
     * @param mode LTO mode
     * @param d  emission value
     */
    public void set(String trc, String mode, double d) {    
      ht.put(trc.toLowerCase()+"-"+mode.toLowerCase(), d);
    }
    
    /**
     * Gets an emission value
     * @param trc tracer name
     * @param mode LTO mode
     * @return emission value or Double.NaN
     */
    public double get(String trc, String mode) {    
      Double d = ht.get(trc.toLowerCase()+"-"+mode.toLowerCase());
      return (d == null) ? Double.NaN : d;
    }
    
    /**
     * Gets the emission over the certification LTO
     * @param trc tracer name
     * @return LTO emission or Double.NaN
     */
    public double getLTOEmission(String trc) {
      return getLTOEmission(trc, true, true);
    }
    
    /**
     * Gets the emission over the certification LTO
     * @param trc tracer name
     * @param arr
     * @param dep
     * @return LTO emission or Double.NaN
     */
    public double getLTOEmission(String trc, boolean arr, boolean dep) {
      double sum = 0;
      for (MOD.MODE m : MOD.MODE.values()) {
        if (!arr && (m == MOD.MODE.AP))
          continue;
        if (!dep && (m == MOD.MODE.TO || m == MOD.MODE.CO))
          continue;
        double tim = MOD.ICAOTIM[m.ordinal()];
        if (m == MOD.MODE.ID) {
          double t = 0;
          if (arr) t += tim*(9.0/26.0);
          if (dep) t += tim*(17.0/26.0);
          tim = t;
        }
        double ff, ei;
        if (!trc.equalsIgnoreCase(EEDB.FUELFLOW))
          ff = get(EEDB.FUELFLOW, m.toString());
        else
          ff = 1.;
        ei = get(trc, m.toString());
        if (ff >= 0 && ei >= 0)
          sum += ff * tim * ei;
        else
          return Double.NaN;
      }
      return sum;      
    }
    
    /**
     * Returns a string representation of the engine data and the emission values
     * for default tracers
     * @return string
     */
    @Override
    public String toString() {
      return toString(EEDB.EEDBTRCLSP, false);
    }
    
    /**
     * Returns a string representation of the engine data and the emission values
     * for the specified tracers
     * @param trc tracer names or empty string array of length zero
     * @return string
     */
    public String toString(String[] trc) {
      return toString(trc, false);
    }

    /**
     * Returns a string representation of the engine data and the emission values
     * for the specified tracers
     * @param trc tracer names or empty string array of length zero
     * @param blto if true, the relative contribution to the LTO modes is also listed
     * @return string
     */
    public String toString(String[] trc, boolean blto) {
      StringBuilder sb = new StringBuilder();
      sb.append("\n");
      sb.append("[PARAMETER.ENGINE.").append(UID).append("]\n");
      sb.append("UID; ").append(String.format("%s", UID)).append("\n");
      sb.append("ENG; ").append(String.format("%s", ENG)).append("\n");
      sb.append("SRC; ").append(String.format("%s", SRC)).append("\n");
      sb.append("ETP; ").append(String.format("%s", ETP)).append("\n");
      sb.append("CMB; ").append(String.format("%s", CMB)).append("\n");
      sb.append("BYP; ").append(String.format("%1.2f", BYP)).append("\n");
      sb.append("PRR; ").append(String.format("%1.2f", PRR)).append("\n");
      sb.append("ROP; ").append(String.format("%1.2f", ROP)).append("\n");
      sb.append("SNM; ").append(String.format("%1.2f", get(EEDB.SMOKENUMBER, EEDB.SNMAX))).append("\n");
      sb.append("FTD; ").append(String.format("%s", FTD)).append("\n");
      sb.append("SSB; ").append(String.format("%b", SSB)).append("\n");
      sb.append("SSU; ").append(String.format("%s", SSU)).append("\n");
      sb.append("SRG");
      if (SRG != null)
        for (String ss : SRG)
          sb.append("; ").append(ss);
      else
        sb.append(";");
      sb.append("\n");
      sb.append("PMM; ").append(String.format("%s", PMM)).append("\n");
      sb.append("\n");
      sb.append("[TABLE.ENGINE.").append(UID).append("]\n");
      sb.append(String.format("%-16s; %-5s; %14s; %14s; %14s; %14s; %14s\n",
        "Name", "Unit", "Takeoff", "Climbout", "Approach", "Idle", "LTO"));
      double[] lto = new double[trc.length];
      int ilto = 0;
      for (String st : trc) {
        String unit = "g/kg";
        if (st.equalsIgnoreCase(EEDB.FUELFLOW))
          unit = "kg/s";
        if (st.equalsIgnoreCase(EEDB.ODOR))
          unit = "OU/kg";
        else if (st.equalsIgnoreCase(EEDB.SMOKENUMBER))
          unit = "1";
        else if (st.startsWith(PM.GMD))
          unit = "nm";
        else if (st.startsWith(PM.CEE))
          unit = "g/m3";
        else if (st.contains("PN"))
          unit = "1/kg";
        sb.append(String.format("%-16s; %-5s", st, unit));
        double sum = 0;
        String form = "";
        for (MOD.MODE m : MOD.MODE.values()) {
          double ff = this.get(EEDB.FUELFLOW, m.toString());
          double tim = MOD.ICAOTIM[m.ordinal()];
          double ei = this.get(st, m.toString());
          if (st.equalsIgnoreCase(EEDB.FUELFLOW))
            ff = 1.;
          if (Double.isNaN(ei))
            ei = -999;
          form = (unit.equals("1/kg") || unit.equals("g/m3") || unit.equals("OU/kg")) ? 
            "; %14.5e" : "; %14.5f";
          sb.append(String.format(form, ei));
          if (ei > 0) sum += tim*ff*ei;
        }
        lto[ilto++] = sum;
        if (st.equalsIgnoreCase(EEDB.SMOKENUMBER) || st.startsWith(PM.GMD) || st.startsWith(PM.CEE))
          sb.append(String.format("; %14s"+"\n", " "));
        else {
          sb.append(String.format(form+"\n", sum));
        }
      }  
      if (!blto)
        return sb.toString();
      ilto = 0;
      for (String st : trc) {
        String unit = "%LTO";
        if (st.equalsIgnoreCase(EEDB.FUELFLOW) || st.equalsIgnoreCase(EEDB.SMOKENUMBER)) {
          ilto++;
          continue;
        }
        sb.append(String.format("%-16s; %-5s", st, unit));
        String form = "; %12.0f";
        for (MOD.MODE m : MOD.MODE.values()) {
          double ff = this.get(EEDB.FUELFLOW, m.toString());
          double tim = MOD.ICAOTIM[m.ordinal()];
          double ei = this.get(st, m.toString());
          if (ei < 0) ei = 0;   
          sb.append(String.format(form, (lto[ilto] > 0) ? ei*tim*ff/lto[ilto]*100 : 0));
        }
        sb.append(String.format(form+"\n", 100.));
        ilto++;
      }  
      return sb.toString();
    }
 
}
