////////////////////////////////////////////////////////////////////////////////
//
// history:
//
// 2022-01-04  uj  0.1.0  first setup
// 2022-04-14  uj  1.0.0  first release
// 2022-04-24  uj  1.0.1  additional security checks
// 2022-07-14  uj  1.1.0  EEDB handle negative modified values in the FOI data
// 2023-03-03  uj  1.1.1  PM10_VH renamed to PM10_VO
// 2023-07-04  uj  1.1.2  EEDB handle empty UID string in XLS EEDB
//
//
////////////////////////////////////////////////////////////////////////////////

/**
 * (c) Janicke Consulting, Germany 2023
 * mail info@janicke.de
 * 
 * Simple Engine emission CalculaTOR.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Version 1.0.0 was developed in the project RAPTOR which has
 * received funding from the Clean Sky 2 Joint Undertaking under
 * the European Union's Horizon 2020 research program under grant
 * agreement No 863969.
 *
 */

package de.janicke.airport.sector;

import de.janicke.airport.eedb.ACT;
import de.janicke.airport.eedb.ACG;
import de.janicke.airport.eedb.CST;
import de.janicke.airport.eedb.EEDB;
import de.janicke.airport.eedb.ENG;
import de.janicke.airport.eedb.MOD;
import de.janicke.airport.eedb.PM;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.net.URI;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;
import java.util.zip.CRC32;

public class Sector {

  public static final String VERSION = "1.1.2";
  public static final String SETTINGS = "Sector.settings";
  public static final String SETTINGS_CRC32 = "1EC9A180";

  public  boolean Verbose;
  private String Dir, Base, Mark, Fnin, Fnout;
  private BufferedWriter bwLog;
  private EEDB eedb;
  private PM.METHOD pm_method;
  private double[] daFOA4USR;
  private String[] saListing;
  private LinkedHashMap<String, MOV> hm;
  private LinkedHashMap<String,ACT> ht;
  private LinkedHashMap<String,String> hfoi;
  private boolean FOIavailable;
  
  /**
   * internal class to store movements/operations
   */
  private class MOV {
    public String acg;
    public String act;
    public String uid;
    public String apu;
    public int nen;
    public double narr;
    public double ndep;
    private String key;
    
    public String getKey() {
      StringBuilder sb = new StringBuilder();
      sb.append((acg != null) ? acg.toLowerCase() : "#1");
      sb.append("-").append((act != null) ? act.toLowerCase() : "#2");
      sb.append("-").append((uid != null) ? uid.toLowerCase() : "#3");
      sb.append("-").append((apu != null) ? apu.toLowerCase() : "#4");
      sb.append("-").append((nen > 0) ? ""+nen : "#5");
      key = sb.toString();
      return key;
    } 
  }
  
  /**
   * put an aircraft type to the internal list
   * @param act 
   */
  private void putACT(ACT act) {
    if (ht == null)
      ht = new LinkedHashMap<String,ACT>();
    if (act == null || act.name == null)
      return;
    ht.put(act.name.toUpperCase(), act);
  }
  
  /**
   * get an aircraft type from the internal list
   * @param name
   * @return 
   */
  private ACT getACT(String name) {
    if (ht == null || name == null)
      return null;
    return ht.get(name.toUpperCase());
  }
  
  /** 
   * put a movement to the internal list
   * @param mov 
   */
  private void putMOV(MOV mov) {
    if (hm == null)
      hm = new LinkedHashMap<String,MOV>();
    if (mov == null)
      return;
    hm.put(mov.getKey(), mov);
  }
  
  /**
   * gets a movement from the inetrnal list
   * @param key
   * @return 
   */
  private MOV getMOV(String key) {
    if (hm == null || key == null)
      return null;
    return hm.get(key);
  }

  /**
   * Main class
   * @param dir  project directory
   * @param mark  project marker
   */
  public Sector(String dir, String mark) {
    Dir = dir;
    Mark = (mark == null) ? "" : mark;
    Verbose = false;
  }
  
  /**
   * init the class
   */
  private void init() {
    Base = my_home();
    Fnin = "sector"+((Mark.length() > 0) ? "-"+Mark+".txt" : ".txt");
    Fnout = "sector"+((Mark.length() > 0) ? "-"+Mark+".log" : ".log");
    eedb = new EEDB(); 
    bwLog = null;
    FOIavailable = false;
    try {
      //
      // start logging
      File flog = new File(Dir, Fnout);
      bwLog = new BufferedWriter(new FileWriter(flog));
      SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'"+"T"+"'HH:mm:ssZ");
      sdf.setTimeZone(TimeZone.getTimeZone("GMT+01:00"));
      log(sdf.format(new Date(Calendar.getInstance().getTimeInMillis()))+
        ", JAVA "+System.getProperty("java.version"));
      log("SECTOR %s (c) 2022-2023 Janicke Consulting, Germany", VERSION);
      log("");
      log("This program is free software; you can redistribute it and/or");
      log("modify it under the terms of the GNU General Public License as");
      log("published by the Free Software Foundation; either version 3 of");
      log("the License, or (at your option) any later version.");
      log("");
      log("This program is distributed in the hope that it will be useful,");
      log("but WITHOUT ANY WARRANTY; without even the implied warranty of");
      log("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU");
      log("General Public License for more details.");
      log("");
      log("Version 1.0.0 was developed in the project RAPTOR which has");
      log("received funding from the Clean Sky 2 Joint Undertaking under");
      log("the European Union's Horizon 2020 research program under grant");
      log("agreement No 863969.");
      log("");
      log(flog.getAbsolutePath());
      //
      // read settings
      if (!(new File(Base, SETTINGS)).exists())
        throw new Exception("settings file "+SETTINGS+" not found!");
      readSettings();
    }
    catch (Exception e) {
      if (bwLog != null)
        log(e.getMessage());
      try {
        bwLog.close();
        bwLog = null;
      }
      catch (Exception ee) {}
    }
  }
  
  /**
   * closes the class
   */
  private void close() {
    try {
      log("");
      log("SECTOR finished.");
      bwLog.close();
    }
    catch (Exception e) {}
  }
  
  /**
   * log information to the log file and to the stdout
   * @param frm
   * @param args 
   */
  private void log(String frm, Object ... args) {
    if (bwLog == null) 
      return;
    try {
      bwLog.write(String.format(frm, args));
      bwLog.newLine();
      if (Verbose) {
        System.out.printf(frm, args);
        System.out.printf("\n");
      }
    }
    catch (Exception e) {}
  }
  
  /**
   * gets a CSV object 
   * @param section
   * @param bdefault  true if from the settings file
   * @return
   * @throws Exception 
   */
  private CST getCs(String section, boolean bdefault) throws Exception {
    CST cs = null;
    if (bdefault)
      cs = new CST(new File(Base, SETTINGS), section, ';');
    else
      cs = new CST(new File(Dir, Fnin), section, ';');
    if (bdefault && cs.getColumnCount() < 0)
      throw new Exception("section "+section+" not found in settings file!");
    return cs;
  }
  
  /**
   * gets the home directory from which the application starts
   * @return 
   */
  private static String my_home() {
    Class<Sector> cls = Sector.class;
    String nm = cls.getName();
    URL u2 = cls.getResource(cls.getSimpleName() + ".class"); 
    String s2;
    File root = null;
    try {
      URI uri = u2.toURI();
      String sScheme = uri.getScheme();
      s2 = uri.getSchemeSpecificPart();
      if (sScheme.equals("jar")) { 
        if (s2.startsWith("file:")) s2 = s2.substring(5); 
      } 
      else if (!sScheme.equals("file")) { 
        return null;
      }
      int k = s2.indexOf(nm.replace('.', '/'));
      if (k > 0) s2 = s2.substring(0, k-1);
      if (s2.endsWith(".jar!")) { 
        s2 = s2.substring(0, s2.length() - 1);
        root = (new File(s2)).getParentFile();
      }
      else {
        root = new File(s2);
      }
      if (root.getName().equals("classes") 
          && root.getParentFile().getName().equals("build"))
        root = root.getParentFile().getParentFile();
    } 
    catch (Exception e) {
      e.printStackTrace(System.out);
    }
    return (root != null) ? root.getAbsolutePath() : null;
  }
  
  /**
   * reads the settings file
   * @throws Exception 
   */
  private void readSettings() throws Exception {
    //
    // check integrity
    CRC32 crc32 = new CRC32();
    File f = new File(Base, SETTINGS);
    FileInputStream fis = new FileInputStream(f);
    byte[] arr = new byte[(int)f.length()];
    fis.read(arr, 0, arr.length);
    fis.close();
    crc32.reset();
    crc32.update(arr);
    String sc = String.format("%x", crc32.getValue());
    if (!sc.equalsIgnoreCase(SETTINGS_CRC32))
      throw new Exception("Modified settings file encountered, CRC32 is "+
      sc.toUpperCase()+", expected "+SETTINGS_CRC32);
    log("Using Sector.settings with CRC %s", sc.toUpperCase());
    //
    String spmm = null, s;
    double[] da = null;
    String[] sa = null;
    daFOA4USR = null;
    saListing = null;
    StringBuilder sb = new StringBuilder();
    for (int im=0; im<2; im++) {
      CST cs = getCs("PARAMETER.SETTINGS", im == 0);
      if (cs.getColumnCount() < 0)
        continue;
      eedb.setPrm("fnEEDB", String.class, cs.getString("Filename_EEDB"));
      eedb.setPrm("fnFOCA", String.class, cs.getString("Filename_FOCA"));
      eedb.setPrm("fnFOI",  String.class, cs.getString("Filename_FOI"));
      eedb.setPrm("EI_CO2", Double.class, cs.getDouble("EI_CO2(g/kg)"));
      eedb.setPrm("EI_FB", Double.class, cs.getDouble("EI_FB(g/kg)"));
      eedb.setPrm("EI_SOX", Double.class, cs.getDouble("EI_SOX(g/kg)"));
      eedb.setPrm("FSC", Double.class, cs.getDouble("FSC(1)"));
      eedb.setPrm("S4TOS6", Double.class, cs.getDouble("S4TOS6(1)"));
      eedb.setPrm("H2O", Double.class, cs.getDouble("H2O(1)"));
      eedb.setPrm("BNZ_IN_HC", Double.class, cs.getDouble("BNZ_IN_HC(g/g)"));
      eedb.setPrm("ODOR_IN_HC", Double.class, cs.getDouble("ODOR_IN_HC(OU/g)"));
      eedb.setPrm("PM25_IN_PM10", Double.class, cs.getDouble("PM25_IN_PM10(g/g)"));
      eedb.setPrm("USE_CERT_LTO", Integer.class, cs.getInt("USE_CERT_LTO"));
      s = cs.getString("PM_Method");
      if (s != null)
        spmm = s;
      da = cs.getDoubles("PM_FOA4USR");
      if (da != null && da.length == 8)
        daFOA4USR = da;
      sa = cs.getStrings("Listing");
      if (sa != null)
        saListing = sa;
    }
    //
    // PM method
    pm_method = PM.METHOD.FOA4;
    if (spmm != null) {
      for (PM.METHOD m : PM.METHOD.values())
        if (m.toString().equalsIgnoreCase(spmm)) {
          pm_method = m;
          break;
        }
    }
    //
    // PM user-defined parms for FOA4USR
    if (daFOA4USR != null) {
      for (int i=0; i<4; i++) 
        daFOA4USR[i] *= 1.e-9;
    }
    //
    log("%-20s: %1.3f", "EI_CO2(g/kg)", eedb.getPrmDouble("EI_CO2"));
    log("%-20s: %1.3f", "EI_FB(g/kg)", eedb.getPrmDouble("EI_FB"));
    log("%-20s: %1.3e", "FSC(1)", eedb.getPrmDouble("FSC"));
    log("%-20s: %1.3e", "S4TOS6(1)", eedb.getPrmDouble("S4TOS6"));
    log("%-20s: %1.3e", "H2O(1)", eedb.getPrmDouble("H2O"));
    log("%-20s: %1.3f", "EI_SOX(g/kg)", eedb.getPrmDouble("EI_SOX"));
    log("%-20s: %1.3f", "BNZ_IN_HC(g/g)", eedb.getPrmDouble("BNZ_IN_HC"));
    log("%-20s: %1.3f", "ODOR_IN_HC(OU/g)", eedb.getPrmDouble("ODOR_IN_HC"));
    log("%-20s: %1.3f", "PM25_IN_PM10(g/g)", eedb.getPrmDouble("PM25_IN_PM10"));
    log("%-20s: %1d", "USE_CERT_LTO", eedb.getPrmInt("USE_CERT_LTO"));
    log("%-20s: %s", "PM_Version", PM.VERSION);
    log("%-20s: %s", "PM_Method", pm_method);
    if (daFOA4USR != null)
      log("%-20s: %1.1f; %1.1f; %1.1f; %1.1f; %1.2f; %1.2f; %1.2f; %1.2f", "PM_FOA4USR",
        daFOA4USR[0]*1.e9, daFOA4USR[1]*1.e9, daFOA4USR[2]*1.e9, daFOA4USR[3]*1.e9,
        daFOA4USR[4], daFOA4USR[5], daFOA4USR[6], daFOA4USR[7]);
    if (saListing != null) {
      sb.setLength(0);
      for (String ss : saListing)
        sb.append((sb.length() > 0 ? "; " : "")).append(ss);
      log("%-20s: %s", "Listing", sb.toString());
    }
  }
  
  /**
   * reads the movements from the input file
   * @throws Exception 
   */
  private void readMovements() throws Exception {
    //
    // read aircraft types
    String name, act, acg, uid, apu;
    int nen;
    ACT a;
    String s;
    int nadded, nreplaced, nremoved;
    boolean newact;
    for (int im=0; im<2; im++) {
      nadded = 0;
      nreplaced = 0;
      nremoved = 0;
      s = (im == 0) ? "default" : "project";
      CST cs = getCs("TABLE.AIRCRAFT.TYPES", im == 0);
      if (cs.getRowCount() < 0)
        continue;
      for (int ir=0; ir<cs.getRowCount(); ir++) {
        name = cs.getString(ir, "ACT");
        if (name == null)
          continue;
        a = getACT(name);
        if (a == null) {
          a = new ACT(name);
          newact = true;
        }
        else
          newact = false;
        acg = cs.getString(ir, "ACG");
        if (acg != null)
          a.acg = acg;
        uid = cs.getString(ir, "UID");
        if (uid != null)
          a.uid = uid;
        apu = cs.getString(ir, "APU");
        if (apu != null)
          a.apu = apu;
        nen = cs.getInt(ir, cs.getColumnIndex("NEN"));
        if (nen != Integer.MIN_VALUE)
          a.nen = nen;
        //
        // check
       // if (!eedb.isValidACG(a.acg) || !eedb.isValidUID(a.uid) || !eedb.isValidNEN(a.nen)) {
        if (!eedb.isValidACG(a.acg) || !eedb.isValidNEN(a.nen)) {
          if (!newact) nremoved++;
          if (!FOIavailable && hfoi != null && a.uid != null && hfoi.get(a.uid.toLowerCase()) != null)
            ;
          else
            log("*** aircraft type %s/%s/%s with invalid ACG/UID/NEN, removed", a.name, a.acg, a.uid);
        }
        else {
          if (newact) {
            putACT(a);
            nadded++;
          }
          else
            nreplaced++;
        }
      }
      log("%-20s: %d/%d/%d aircraft types added/modified/removed from %s settings",  
        "ACT", nadded, nreplaced, nremoved, s);
    }
    //
    // read movements
    CST cs = getCs("TABLE.MOVEMENTS", false);
    double narr = 0;
    double ndep = 0;
    int iacg = cs.getColumnIndex("ACG");
    int iact = cs.getColumnIndex("ACT");
    int iuid = cs.getColumnIndex("UID");
    int inen = cs.getColumnIndex("NEN");
    int iapu = cs.getColumnIndex("APU");
    int iadl, iops;
    iadl = cs.getColumnIndex("A/D/L");
    if (iadl < 0)
      iadl = cs.getColumnIndex("A/D");
    iops = cs.getColumnIndex("OPS");
    if (iadl < 0 && iops < 0)
       iops = cs.getColumnIndex("LTO");
    double ops;
    String type;
    for (int ir=0; ir<cs.getRowCount(); ir++) {
      acg = cs.getString(ir, iacg);
      act = cs.getString(ir, iact);
      uid = cs.getString(ir, iuid);
      apu = cs.getString(ir, iapu);
      nen = cs.getInt(ir, inen);
      //
      // ACG/ACT
      if (!eedb.isValidACG(acg) && getACT(act) != null)
        acg = getACT(act).acg;
      if (!eedb.isValidACG(acg)) {
        log(String.format("*** row %d: can't assign valid ACG (ACT %s), skipped", ir+1, act));
        continue;
      }
      if (getACT(act) == null)
        act = null;
      if (uid != null && uid.length() == 0)
        uid = null;
      //
      // replace by ACT parameters if required and possible
      if (act != null) {
        a = getACT(act);
        if (uid == null)
          uid = a.uid;
        else if (!eedb.isValidUID(uid))
          uid = a.uid;
        if (!eedb.isValidNEN(nen))
          nen = a.nen;
        if (apu == null)
          apu = a.apu;
      }
      else if (uid != null) {
        if (!eedb.isValidUID(uid) || !eedb.isValidNEN(nen)) {
          log(String.format("*** row %d: can't assign valid UID/NEN ("+uid+"/"+nen+"), setting ignored", ir+1));
          uid = null;
          nen = -1;
        }
      }
      if (!eedb.isValidAPU(apu))
        apu = eedb.getACG(acg).APU;
      // at this point all movement entries should have a valid
      // ACG and optionally a valid UID and NEN
      ops = cs.getDouble(ir, iops);
      if (iops < 0)
        ops = 1.;
      if (Double.isNaN(ops) || ops < 0) {
        log(String.format("*** row %d: can't assign operations, skipped", ir+1));
        continue;
      }
      MOV m = new MOV();
      m.acg = acg;
      m.act = act;
      m.uid = uid;
      m.nen = nen;
      m.apu = apu;
      type = cs.getString(ir, iadl);
      if (type == null || type.equalsIgnoreCase("L")) {
        m.narr = ops;
        m.ndep = ops;
        narr += ops;
        ndep += ops;
      }
      else if (type.equalsIgnoreCase("A")) {
        m.narr = ops;
        m.ndep = 0;
        narr += ops;
      }
      else if (type.equalsIgnoreCase("D")) {
        m.narr = 0;
        m.ndep = ops;
        ndep += ops;
      }
      else {
        log(String.format("*** row %d: can't assign operation type, skipped", ir+1));
        continue;
      }
      MOV mm = getMOV(m.getKey());
      if (mm != null) {
        mm.narr += m.narr;
        mm.ndep += m.ndep;
      }
      else 
        putMOV(m);
    }
    if (hm == null)
      log("%-20s: no operations defined!", "MOV");
    else 
      log("%-20s: %d effective movements with %1.1f arr, %1.1f dep, %1.1f LTO", "MOV", 
        hm.size(), narr, ndep, (narr+ndep)/2.);
  }
  
  /**
   * reads the emission data: first the original XLS files. Then more data, 
   * first from the settings file, then from the input file.
   * @throws Exception 
   */
  private void readEmissions() throws Exception {
    eedb.CHECK = false;
    int n = 0;
    int[] na;
    File f;
    //
    // read EEDB engines
    f = new File(Base, eedb.getPrmString("fnEEDB"));
    n = eedb.readEEDB(f.getAbsolutePath(), "EEDB");
    log("%-20s: %d engines from %s", "ENG EEDB", n, f.getName());
    //
    // read FOCA engines
    f = new File(Base, eedb.getPrmString("fnFOCA"));
    n = eedb.readFOCA(f.getAbsolutePath(), "FOCA");
    log("%-20s: %d engines from %s", "ENG FOCA", n, f.getName());
    //
    // read FOI engines
    f = new File(Base, eedb.getPrmString("fnFOI"));
    try {
      n = eedb.readFOI(f.getAbsolutePath(), "FOI");
      FOIavailable = true;
    }
    catch (Exception e) {
      n = -1;
    }
    if (n < 0)
      log("%-20s: not provided", "ENG FOI");
    else
      log("%-20s: %d engines from %s", "ENG FOI", n, f.getName());
    //
    // read CSV engines
    for (int im=0; im<2; im++) {
      f = (im == 0) ? new File(Base, SETTINGS) : new File(Dir, Fnin);
      String s = (im == 0) ? "default" : "project";
      String ss = (im == 0) ? "SECTOR" : Mark;
      boolean check = (im == 0);
      na = eedb.readCSV(f.getAbsolutePath(), "TABLE.EEDB.OTHER", EEDB.CSVTRC, ss, check);
      if (na != null && (na[0] > 0 || na[1] > 0))
        log("%-20s: %d/%d engines added/modified from %s settings", "ENG CSV", 
          na[0], na[1], s);
    }
    //
    // read CSV group emissions
    for (int im=0; im<2; im++) {
      f = (im == 0) ? new File(Base, SETTINGS) : new File(Dir, Fnin);
      String s = (im == 0) ? "default" : "project";
      String ss = (im == 0) ? "SECTOR" : Mark;
      boolean check = false;
      na = eedb.readCSV(f.getAbsolutePath(), "TABLE.EEDB.ACG", EEDB.CSVTRC, ss, check);
      if (na != null && (na[0] > 0 || na[1] > 0))
        log("%-20s: %d/%d aircraft groups (EEDB) added/modified from %s settings", "ACG CSV", na[0], na[1], s);
    }
    //
    // read APU
    for (int im=0; im<2; im++) {
      f = (im == 0) ? new File(Base, SETTINGS) : new File(Dir, Fnin);
      String s1 = (im == 0) ? "default" : "project";
      String s2 = (im == 0) ? "added" : "modified";
      String ss = (im == 0) ? "SECTOR" : Mark;
      n = eedb.readAPU(f.getAbsolutePath(), "TABLE.EEDB.APU", EEDB.CSVTRC, ss);
      if (n > 0)
        log("%-20s: %d APU %s from %s settings", "APU CSV", n, s2, s1);
    }
    //
    // read TIM
    for (int im=0; im<2; im++) {
      f = (im == 0) ? new File(Base, SETTINGS) : new File(Dir, Fnin);
      String s1 = (im == 0) ? "default" : "project";
      String s2 = (im == 0) ? "added" : "modified";
      n = eedb.readTIM(f.getAbsolutePath(), "TABLE.LTO.SECONDS", 
        eedb.getPrmInt("USE_CERT_LTO") != 0);
      if (n > 0)
        log("%-20s: %d aircraft groups (TIM) %s from %s settings", 
        "TIM CSV", n, s2, s1);  
    }
    //
    // insert FOI UIDs
    f = new File(Base, SETTINGS);
    CST css = new CST(f, "TABLE.UID.MAPPING", ';');
    ArrayList<String> va = eedb.getUIDs();
    for (String s : va) {
      ENG eng = eedb.get(s);
      if (eng.ISRC != EEDB.SOURCE.FOI.hashCode())
        continue;
      boolean inserted = false;
      for (int ir=0; ir<css.getRowCount(); ir++) {
        String seng = css.getString(ir, "ENG");
        if (eng.ENG.toLowerCase().equals(seng.toLowerCase())) {
          eedb.remove(eng.UID);
          eng.UID = css.getString(ir, "UID");
          eedb.put(eng);
          inserted = true;
          break;
        }
      }
      if (!inserted) {
        log("*** no UID set for FOI engine %s, removed", eng.ENG);
        eedb.remove(eng.UID);
      }
    }
    //
    // insert surrogates and make a list of FOI UIDs
    if (hfoi == null)
      hfoi = new LinkedHashMap<String,String>();
    else
      hfoi.clear();
    va = eedb.getUIDs();
    for (int ir=0; ir<css.getRowCount(); ir++) {
      String src = css.getString(ir, "SRC");
      String srg = css.getString(ir, "SRG");
      String uid = css.getString(ir, "UID");
      if (src.equalsIgnoreCase("FOI") && uid.length() > 0)
        hfoi.put(uid.toLowerCase(), uid);
      if (srg.length() == 0)
        continue;
      String[] sa = srg.split("[/]");
      ENG e = eedb.get(uid);
      if (e != null) {
        e.SRG = sa;
      }
      else {
        for (String s : va) {
          e = eedb.get(s);
          if (e.ENG != null && e.ENG.equalsIgnoreCase(css.getString(ir, "ENG"))) {
            e.SRG = sa;
          }
        }
      }
    }
  }

  /**
   * inserts the PM EIs for all methods, then copies the selected methods
   * @throws Exception 
   */
  private void insertEmissions() throws Exception {
    eedb.insertPM(daFOA4USR, false); // line-loss corrected
    eedb.insertPM(daFOA4USR, true);  // uncorrected
    eedb.copyMethod(pm_method, false, false); // copy corrected method
    eedb.insertOther();
    log("%-20s: emission indices and PM methods inserted", "EI/PM");
  }
  
  /**
   * writes the user-specified listings to the output file
   * @throws Exception 
   */
  private void writeListings() throws Exception {
    if (saListing == null)
      saListing = new String[0];
    int mass_level = 1;
    for (String s : saListing) {
      int i = s.lastIndexOf("-");
      int level = (i < 0) ? 1 : Integer.parseInt(s.substring(i+1)); 
      if (i > 0)
        s = s.substring(0,i);
      s = s.toUpperCase();
      if (s.equals("EEDB")) {
        eedb.writeEEDB(bwLog, getOutputColumns(level));
      }
      else if (s.equals("APU")) {
        int l = level;
        if (l > 2)
          l = 2;
        eedb.writeAPU(bwLog, getOutputColumns(l));
      }
      else if (s.equals("EMIS")) {
        mass_level = level;
        if (mass_level > 2)
          mass_level = 2;
      }
      else if (s.equals("ACGX")) {
        writeEI(bwLog, level, true);
      }
      else if (s.equals("ACG")) {
        writeEI(bwLog, level, false);
      }
      else if (eedb.get(s) != null) {
        log(eedb.get(s).toString(getOutputColumns(level)));
      }
      else if (s.equals("MOV"))
        writeMOV(bwLog);
      else if (s.equals("ACT"))
        writeACT(bwLog);
      else if (s.startsWith("VPM@"))
        writeVPM(bwLog, eedb.get(s.substring(4)), level);
    }
    eedb.writeTIM(bwLog);
    writeMass(bwLog, mass_level);
  }
  
  /**
   * calculates the emitted masses
   * @throws Exception 
   */
  private void calculateMass() throws Exception {
    if (hm == null)
      return;
    String[] trca = getOutputColumns(2);
    ACG gtt = new ACG(EEDB.TOTAL);
    eedb.putACG(gtt);
    //
    // reset
    for (EEDB.ACGT g : EEDB.ACGT.values()) {
      ACG acg = eedb.getACG(g.toString());
      for (String t : trca) {
        for (MOD.MODE m : MOD.MODE.values())
          acg.setMass(t, m.toString(), 0.);
        acg.setMass(t, EEDB.APU, 0.);
        acg.narr = 0.;
        acg.ndep = 0.;
      }
    }
    //
    // calculate
    double ff, ei, tim, n, mass;
    String sm;
    for (String key : hm.keySet()) {
      MOV mov = getMOV(key);
      ACG acg = eedb.getACG(mov.acg);
      acg.narr += mov.narr;
      acg.ndep += mov.ndep;
      gtt.narr += mov.narr;
      gtt.ndep += mov.ndep;
      boolean use_group_default = false;
      ENG eng = (mov.uid != null) ? eedb.get(mov.uid) : null;
      if (eng == null) {
        eng = eedb.get(mov.acg);
        use_group_default = true;
        if (mov.uid != null)
          log("%-30s: engine not found, using group default", key);
      }
      for (String t : trca) {
        if (t.equalsIgnoreCase(EEDB.FUELFLOW)) continue;
        if (t.equalsIgnoreCase(EEDB.SMOKENUMBER)) continue;       //-2022-04-24
        if (use_group_default) {
          if (t.equalsIgnoreCase(EEDB.PM10_NV) || 
            t.equalsIgnoreCase(EEDB.PM10_VS) ||
            t.equalsIgnoreCase(EEDB.PM10_SO4) ||
            t.equalsIgnoreCase(EEDB.PM10_POM) ||
            t.equalsIgnoreCase(EEDB.PM10_VO) ||
            t.equalsIgnoreCase(EEDB.PM10_VH))
          continue;
        }
        //
        // main engines
        for (MOD.MODE m : MOD.MODE.values()) {
          sm = m.toString();
          ff = eng.get(EEDB.FUELFLOW, sm);
          //
          if (!use_group_default) 
            ff *= mov.nen;
          //
          ei = eng.get(t, sm);
          tim = acg.getTIM(sm);
          n = 0.;
          if (m == MOD.MODE.AP)
            n = mov.narr;
          else if (m == MOD.MODE.ID)
            n = (9.*mov.narr + 17.*mov.ndep)/26.; 
          else if (m == MOD.MODE.CO || m == MOD.MODE.TO)
            n = mov.ndep;
          //
          // check just in case for unexpected invalid values        2022-04-24
          if (Double.isNaN(ff) || ff < 0.)
            throw new Exception(String.format("*** mov %s/%s: invalid fuel flow encountered!", mov.key, sm));
          if (!use_group_default && (mov.nen > 10 || mov.nen < 0))
            throw new Exception(String.format("*** mov %s/%s: invalid number of engines encountered!", mov.key, sm));
          if (Double.isNaN(ei) || ei < 0.)
            throw new Exception(String.format("*** mov %s/%s/%s: invalid EI encountered!", mov.key, t, sm));
          if (Double.isNaN(tim) || tim < 0.)
            throw new Exception(String.format("*** mov %s/%s: invalid tim encountered!", mov.key, sm)); 
          if (Double.isNaN(n) || n < 0.)
            throw new Exception(String.format("*** mov %s/%s: invalid number of movements encountered!", mov.key, sm));
          //
          mass = ff * ei * tim * n;
          //
          // store tim*n for later extraction of effective group fuel flows
          if (t.equalsIgnoreCase(EEDB.FUELBURN))
            acg.addMass(EEDB.FFWEIGHT, sm, tim * n);
          //
          // store contribution
          acg.addMass(t, sm, mass);
          acg.addMass(t, EEDB.AC, mass);
          gtt.addMass(t, sm, mass);
          gtt.addMass(t, EEDB.AC, mass);
          gtt.addMass(t, EEDB.TOTAL, mass);
        }
        //
        // APU
        if (t.equalsIgnoreCase(EEDB.PM10_NV) || 
            t.equalsIgnoreCase(EEDB.PM10_VS) ||
            t.equalsIgnoreCase(EEDB.PM10_VO) ||
            t.equalsIgnoreCase(EEDB.PM10_SO4) ||
            t.equalsIgnoreCase(EEDB.PM10_POM) ||
            t.equalsIgnoreCase(EEDB.PM10_VH))
          continue;
        if (mov.apu == null || mov.apu.length() == 0)
          continue;
        for (int im=0; im<2; im++) {  // 0: arr, 1: dep
          for (EEDB.APUM m : EEDB.APUM.values()) {
            if (m == EEDB.APUM.HL && im == 0)
              continue;
            sm = m.toString();
            tim = acg.getTIM(sm);
            if (m == EEDB.APUM.SS) {
              tim *= 0.5;
              if (im == 0 && acg.getTIM(EEDB.APUM.NR.toString()) == 0.)
                tim = 0.;
            }
            if (m == EEDB.APUM.NR) {
              tim *= (im == 0.) ? 1./3. : 2./3.;
            }
            if (tim == 0)
              continue;
            ff = eedb.getAPU(mov.apu, EEDB.FUELFLOW, m);
            ff /= 3600.;
            ei = eedb.getAPU(mov.apu, t, m);
            n = (im == 0) ? mov.narr : mov.ndep;
            //
            // check just in case for unexpected invalid values      2022-04-24
            if (Double.isNaN(ff) || ff < 0.)
              throw new Exception(String.format("*** APU %s/%s: invalid fuel flow encountered!", mov.apu, sm));   
            if (Double.isNaN(ei) || ei < 0.)
              throw new Exception(String.format("*** APU %s/%s/%s: invalid EI encountered!", mov.apu, t, sm));
            if (Double.isNaN(tim) || tim < 0.)
              throw new Exception(String.format("*** APU %s/%s: invalid tim encountered!", mov.apu, sm)); 
            if (Double.isNaN(n) || n < 0.)
              throw new Exception(String.format("*** APU %s/%s: invalid number of movements encountered!", mov.apu, sm));
            //
            mass = ff * ei * tim * n;
            //
            // store tim*n for later extraction of effective group fuel flows
            if (t.equalsIgnoreCase(EEDB.FUELBURN))
              acg.addMass(EEDB.FFWEIGHT, sm, tim * n);
            //
            // store contributions
            acg.addMass(t, m.toString(), mass);
            acg.addMass(t, EEDB.APU, mass);
            gtt.addMass(t, EEDB.APU, mass);
            gtt.addMass(t, EEDB.TOTAL, mass);
          }
        } 
      }
    }
  }
  
  
  /**
   * gets the EI for a specific aircraft group
   * @param acg    the aircraft group
   * @param mode   the LTO mode
   * @param t      the tracer
   * @return       fuel flow: g/s, other: g/g
   */
  private double getEI(ACG acg, String mode, String t) {
    double fm = acg.getMass(EEDB.FUELBURN, mode);
    double wg = acg.getMass(EEDB.FFWEIGHT, mode);
    double em = acg.getMass(t, mode);
    if (Double.isNaN(wg) || wg <= 0.)
      return Double.NaN;   
    if (t.equalsIgnoreCase(EEDB.FUELFLOW)) 
      return fm/wg;
    else
      return (fm != 0.) ? em/fm : 0.;
  }
  
  /**
   * writes the aircraft group averaged EIs
   * @param bw
   * @param level
   * @param modx
   * @throws Exception 
   */
  private void writeEI(BufferedWriter bw, int level, boolean modx) throws Exception {
    if (level < 0) level = 0;
    if (level > 2) level = 2;
    String[] trca = getOutputColumns(level);
    StringBuilder sb = new StringBuilder();
    //
    // main engines
    bw.newLine();
    bw.write("[TABLE.EMISSIONS.AC]");
    bw.newLine();
    sb.setLength(0);
    sb.append(String.format("%-12s", "Name"));
    sb.append(String.format("; %-8s", "Tracer"));
    sb.append(String.format("; %-8s", "Unit"));
    if (modx)
      for (MOD.MODX m : MOD.MODX.values())
        sb.append(String.format("; %12s", m));
    else
      for (MOD.MODE m : MOD.MODE.values())
        sb.append(String.format("; %12s", m));
    bw.write(sb.toString());
    bw.newLine();
    String unit, mode;
    double fac, d;
    for (String t : trca) {
      for (EEDB.ACGT g : EEDB.ACGT.values()) {
        sb.setLength(0);
        sb.append(String.format("%-12s", g));
        sb.append(String.format("; %-8s", t));
        unit = "g/kg";
        fac = 1000.;
        if (t.equalsIgnoreCase(EEDB.FUELFLOW)) {
          unit = "kg/s";
          fac = 1./eedb.getPrmDouble("EI_FB");
        }
        else if (EEDB.IsNumberTracer(t))
          unit = "1/kg";
        else if (EEDB.IsOdorTracer(t))
          unit = "OU/kg";
        sb.append(String.format("; %-8s", unit));
        ACG acg = eedb.getACG(g.toString());
        if (modx) {
          for (MOD.MODX m : MOD.MODX.values()) {
            mode = MOD.LSP2LTO(m).toString();
            d = getEI(acg, mode, t);
            if (Double.isNaN(d))
              sb.append(String.format("; %12.0f", -999.));
            else
              sb.append(String.format("; %12.5e", fac*d));
          }
        }
        else {
          for (MOD.MODE m : MOD.MODE.values()) {
            d = getEI(acg, m.toString(), t);
            if (Double.isNaN(d))
              sb.append(String.format("; %12.0f", -999.));
            else
              sb.append(String.format("; %12.5e", fac*d));
          }
        }
        bw.write(sb.toString());
        bw.newLine();
      } 
    }
    //
    // APU
    bw.newLine();
    bw.write("[TABLE.EMISSIONS.APU]");
    bw.newLine();
    sb.setLength(0);
    sb.append(String.format("%-12s", "Name"));
    sb.append(String.format("; %-8s", "Tracer"));
    sb.append(String.format("; %-8s", "Unit"));
    for (EEDB.APUM m : EEDB.APUM.values())
      sb.append(String.format("; %12s", m));
    bw.write(sb.toString());
    bw.newLine();
    for (String t : trca) {
      for (EEDB.ACGT g : EEDB.ACGT.values()) {
        sb.setLength(0);
        sb.append(String.format("%-12s", g));
        sb.append(String.format("; %-8s", t));
        unit = "g/kg";
        fac = 1000.;
        if (t.equalsIgnoreCase(EEDB.FUELFLOW)) {
          unit = "kg/h";
          fac = 3.6;
        }
        else if (EEDB.IsNumberTracer(t))
          unit = "1/kg";
        else if (EEDB.IsOdorTracer(t))
          unit = "OU/kg";
        sb.append(String.format("; %-8s", unit));
        ACG acg = eedb.getACG(g.toString());
        for (EEDB.APUM m : EEDB.APUM.values()) {
          d = getEI(acg, m.toString(), t);
          if (Double.isNaN(d))
            sb.append(String.format("; %12.0f", -999.));
          else
            sb.append(String.format("; %12.5e", fac*d));
        }
        bw.write(sb.toString());
        bw.newLine();
      } 
    } 
  }
  
  /**
   * writes the calculated masses 
   * @param bw
   * @param level
   * @throws Exception 
   */
  private void writeMass(BufferedWriter bw, int level) throws Exception {
    if (hm == null)
      return;
    if (level < 0) level = 0;
    if (level > 2) level = 2;
    String[] trca = getOutputColumns(level, true);
    StringBuilder sb = new StringBuilder();
    String sheader1, sheader2;
    sb.setLength(0);
    sb.append(String.format("%-12s", "Name"));
    for (String t : trca)
      sb.append(String.format("; %12s", t));
    sheader1 = sb.toString();
    sb.setLength(0);
    sb.append(String.format("%-12s", "Unit"));
    for (String t : trca) {
      String s = "Mg";
      if (EEDB.IsNumberTracer(t))
        s = "1";
      else if (EEDB.IsOdorTracer(t))
        s = "MOU";
      sb.append(String.format("; %12s", s));
    }
    sheader2 = sb.toString();
    ArrayList<ACG> vg = new ArrayList<ACG>();
    for (EEDB.ACGT g : EEDB.ACGT.values())
      vg.add(eedb.getACG(g.toString()));
    ACG gtt = eedb.getACG(EEDB.TOTAL);
    vg.add(gtt);
    //
    // movements
    bw.newLine();
    bw.write("[TABLE.MOVEMENTS.SUMMARY]"); bw.newLine();
    bw.write(String.format("%-12s; %10s; %10s; %10s; %10s", 
      "Name", "Arrivals", "Departures", "LTO", "Percent"));
    bw.newLine();
    double ntot = gtt.narr+gtt.ndep;
    for (ACG g : vg) {
      bw.write(String.format("%-12s; %10.1f; %10.1f; %10.1f; %10.1f", 
      g.name, g.narr, g.ndep, 0.5*(g.narr+g.ndep),
      (ntot > 0.) ? 100.*(g.narr+g.ndep)/ntot : 0.));
      bw.newLine();
    }
    bw.newLine();
    //
    // LTO segments
    for (MOD.MODE m : MOD.MODE.values()) {
      bw.write("[TABLE.MASS.AC."+m+"]"); bw.newLine();
      bw.write(sheader1); bw.newLine();
      bw.write(sheader2); bw.newLine();
      for (ACG g : vg) {
        sb.setLength(0);
        sb.append(String.format("%-12s", g.name));
        for (String t : trca) {
          double fac = (EEDB.IsNumberTracer(t)) ? 1. : 1.e-6;
          sb.append(String.format("; %12.5e", fac * g.getMass(t, m.toString())));  
        }
        bw.write(sb.toString()); bw.newLine();
      }
      bw.newLine();
    }
    //
    // total LTO
    bw.write("[TABLE.MASS.AC]"); bw.newLine();
    bw.write(sheader1); bw.newLine();
    bw.write(sheader2); bw.newLine();
    for (ACG g : vg) {
      sb.setLength(0);
      sb.append(String.format("%-12s", g.name));
      for (String t : trca) {
        double fac = (EEDB.IsNumberTracer(t)) ? 1. : 1.e-6;
        sb.append(String.format("; %12.5e", fac * g.getMass(t, EEDB.AC)));  
      }
      bw.write(sb.toString()); bw.newLine();
    }
    bw.newLine();
    //
    // APU
    bw.write("[TABLE.MASS.APU]"); bw.newLine();
    bw.write(sheader1); bw.newLine();
    bw.write(sheader2); bw.newLine();
    for (ACG g : vg) {
      sb.setLength(0);
      sb.append(String.format("%-12s", g.name));
      for (String t : trca) {
        double fac = (EEDB.IsNumberTracer(t)) ? 1. : 1.e-6;
        sb.append(String.format("; %12.5e", fac * g.getMass(t, EEDB.APU)));  
      }
      bw.write(sb.toString()); bw.newLine();
    }
    bw.newLine();
    //
    // total
    bw.write("[TABLE.MASS.TOTAL]"); bw.newLine();
    bw.write(sheader1); bw.newLine();
    bw.write(sheader2); bw.newLine();
    ACG g = eedb.getACG(EEDB.TOTAL);
    sb.setLength(0);
    sb.append(String.format("%-12s", g.name));
    for (String t : trca) {
      double fac = (EEDB.IsNumberTracer(t)) ? 1. : 1.e-6;
      sb.append(String.format("; %12.5e", fac * g.getMass(t, EEDB.TOTAL)));  
    }
    bw.write(sb.toString()); bw.newLine();
  }
  
  /**
   * writes the aircraft types
   * @param bw
   * @throws Exception 
   */
  private void writeACT(BufferedWriter bw) throws Exception {
    bw.write("[TABLE.AIRCRAFT.TYPES]"); bw.newLine();
    bw.write(String.format("%-10s; %-15s; %-15s; %5s; %-5s", "ACT", "ACG", "UID", "NEN", "APU"));
    bw.newLine();
    for (String s : ht.keySet()) {
      ACT act = getACT(s);
      bw.write(String.format("%-10s; %-15s; %-15s; %5d; %-5s", 
        act.name, act.acg, act.uid.toUpperCase(), act.nen, 
        (act.apu != null) ? act.apu : ""));
      bw.newLine(); 
    }
  }
  
    
   /** 
   * writes the movements
   * @param bw
   * @throws Exception 
   */
  private void writeVPM(BufferedWriter bw, ENG eng, int level) throws Exception {
    if (eng == null)
      return;
    bw.newLine();
    ArrayList<String> vtrc = new ArrayList<String>();
    String[] sa = getOutputColumns(level);
    for (String s : sa)
      vtrc.add(s);
    vtrc.add(EEDB.VPM);
    vtrc.add(EEDB.VPN);
    vtrc.add(EEDB.PMR);
    //
    double Dc = 2.e-9; // initial diameter of condensation nuclei
    double fsn = 0.005;  // mass fraction of SO4 in nuclei
    double frn = 0.5;  // mass fraction of remaining reservoir on nuclei
    double Tc = 600.; // time constant for coating
    double gsd = 1.8;
    double m2n = 1./( (Math.PI/6.)*Dc*Dc*Dc*1.e6*Math.exp(4.5*Math.log(gsd)*Math.log(gsd)) );
    //
    for (MOD.MODE m : MOD.MODE.values()) {
      double ei_vs = eng.get(EEDB.PM10_SO4, m.toString());
      eng.set(EEDB.VPM, m.toString(), fsn * ei_vs);
      eng.set(EEDB.VPN, m.toString(), fsn * ei_vs * m2n);  
    }
    sa = new String[vtrc.size()];
    for (int i=0; i<sa.length; i++)
      sa[i] = vtrc.get(i);
    
    bw.write(eng.toString(sa));
    bw.newLine();
    bw.write("====================== chemics.def");
    bw.newLine();
    bw.write(".");
    bw.newLine();
    bw.write(String.format("! %15s | %12s", "created/from", "PMR"));
    bw.newLine();
    bw.write(String.format("C %15s | %12.2e", "NVPM", (1.-frn)/Tc));
    bw.newLine();
    bw.write(String.format("C %15s | %12.2e", "VPM", frn/Tc));
    bw.newLine();
    bw.write(String.format("C %15s | %12.2e", "PMR", -1./Tc));
    bw.newLine();
  }
  
  /** 
   * writes the movements
   * @param bw
   * @throws Exception 
   */
  private void writeMOV(BufferedWriter bw) throws Exception {
    if (hm == null)
      return;
    bw.newLine();
    bw.write("[TABLE.MOVEMENTS]"); bw.newLine();
    bw.write(String.format("%-12s; %-10s; %-15s; %5s; %-5s; %-5s; %10s", 
      "ACG", "ACT", "UID", "NEN", "APU", "A/D/L", "OPS"));
    bw.newLine();
    List<String> keys = new ArrayList<String>(hm.keySet());
    Collections.sort(keys);
    for (String key : keys) {
      MOV m = getMOV(key);
      String type = "L";
      double n = (m.narr + m.ndep)/2.;
      if (m.narr == 0 && m.ndep > 0) {
        type = "D";
        n = m.ndep;
      }
      else if (m.narr > 0 && m.ndep == 0) {
        type = "A";
        n = m.narr;
      }
      bw.write(String.format("%-12s; %-10s; %-15s; %5s; %-5s; %-5s; %10.1f", 
        (m.acg != null) ? m.acg : "", 
        (m.act != null) ? m.act : "", 
        (m.uid != null) ? m.uid : "", 
        (m.nen > 0) ? ""+m.nen : "", 
        (m.apu != null) ? m.apu : "", 
        type,
        n));
      bw.newLine(); 
    }
  }
  
  /**
   * gets the output columns (tracers etc.) for a specified log level
   * @param level
   * @return 
   */
  private String[] getOutputColumns(int level) {
    return getOutputColumns(level, false);
  }
  
  /**
   * gets the output columns (tracers etc.) for a specified log level
   * @param level
   * @param mass true if output table is the calculated mass
   * @return 
   */
  private String[] getOutputColumns(int level, boolean mass) {
    if (level < 0 || level > 4)
      level = 1;
    //
    // level 0: ICAO substances
    // level 1: ICAO plus derived EIs
    // level 2: PM adds and BNZ,ODOR
    // level 3: all PM adds
    // level 4: also uncorrected PM values
    ArrayList<String> vtrc = new ArrayList<String>();
    vtrc.add((mass) ? EEDB.FUELBURN : EEDB.FUELFLOW);
    vtrc.add(EEDB.NOX);
    vtrc.add(EEDB.CO);
    vtrc.add(EEDB.HC);
    if (level > 1) {
      vtrc.add(EEDB.BNZ);
      vtrc.add(EEDB.ODOR);
    }
    if (level > 0) {
      if (!mass) vtrc.add(EEDB.FUELBURN);
      vtrc.add(EEDB.CO2);
      vtrc.add(EEDB.SOX);
      if (level > 1)
         vtrc.add(EEDB.SMOKENUMBER);
      vtrc.add(EEDB.NVPM);
      vtrc.add(EEDB.NVPN);
      if (level > 1) {
        vtrc.add(EEDB.PM10_NV);
        vtrc.add(EEDB.PM10_SO4);
        vtrc.add(EEDB.PM10_POM);
      }
      vtrc.add(EEDB.PM10);
      vtrc.add(EEDB.PM25);
    }
    if (level > 2) { 
      PM.METHOD[] ma = PM.METHOD.values();
      for (PM.METHOD m : ma)
        vtrc.add(EEDB.GMD+"_"+m);
      for (PM.METHOD m : ma)
        vtrc.add(EEDB.CEE+"_"+m);
      for (PM.METHOD m : ma)
        vtrc.add(EEDB.NVPM+"_"+m);
      for (PM.METHOD m : ma)
        vtrc.add(EEDB.NVPN+"_"+m);
    }
    if (level > 3) { 
      vtrc.add(EEDB.NVPM_UC);  
      vtrc.add(EEDB.NVPN_UC);
      PM.METHOD[] ma = PM.METHOD.values();
      for (PM.METHOD m : ma)
        vtrc.add(EEDB.GMD+EEDB.X_UNCORRECTED+"_"+m);
      for (PM.METHOD m : ma)
        vtrc.add(EEDB.CEE+EEDB.X_UNCORRECTED+"_"+m);
      for (PM.METHOD m : ma)
        vtrc.add(EEDB.NVPM+EEDB.X_UNCORRECTED+"_"+m);
      for (PM.METHOD m : ma)
        vtrc.add(EEDB.NVPN+EEDB.X_UNCORRECTED+"_"+m);
    }
    String[] sa = new String[vtrc.size()];
    for (int i=0; i<sa.length; i++)
      sa[i] = vtrc.get(i);
    return sa;
  }
  
  /**
   * runs the calculation
   * @throws Exception 
   */
  public void run() throws Exception {
    Verbose = true;
    init();
    if (bwLog == null)
      throw new Exception();
    readEmissions();
    insertEmissions();
    readMovements();
    calculateMass();
    writeListings();
    close();
  }
  
  public static void main(String[] args) {
    Locale.setDefault(Locale.ENGLISH);
    Sector sec = null;
    try {
      if (args.length != 2) {
        System.out.printf("SECTOR %s\n", VERSION);
        System.out.printf("usage: java -jar jar/Sector.jar <dir> <id>\n");
        System.out.printf("<dir>  input directory\n");
        System.out.printf("<id>   project ID, expected input file is sector-<id>.txt\n");
        System.exit(0);
      }
      sec = new Sector(args[0], args[1]);
      sec.run();
    }
    catch (Exception e) {
      e.printStackTrace(System.out);
    }
    System.exit(0);
  }
  
}
